"""
Handles the reading and processing of any tabular files, as long as the first line
contains the column headers. The columns of the table must be separated using a
separator such as ``,``, ``;``, or ``\\t``.

.. note::

    By default, the second line of the file should contain the units. Alternatively,
    the units can be supplied using extractor parameters, in which case the second line
    is considered to be data.

Since ``yadg-5.0``, the **basic.csv** extractor handles sparse tables (i.e. tables with
missing data) by back-filling empty cells with ``np.NaNs``.

The **basic.csv** extractor attempts to deduce the timestamps from the column headers,
using :func:`yadg.dgutils.dateutils.infer_timestamp_from`. Alternatively, the column(s)
containing the timestamp data and their format can be provided using extractor
parameters.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Basic_csv

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float               # Unix timestamp
      data_vars:
        {{ headers }}:  (uts)                 # Populated from file headers

Metadata
````````
No metadata is extracted.

.. codeauthor::
    Peter Kraus

"""

import logging
from pydantic import BaseModel
from babel.numbers import parse_decimal
from datatree import DataTree
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
from typing import Callable


from yadg import dgutils


logger = logging.getLogger(__name__)


def process_row(
    headers: list,
    items: list,
    datefunc: Callable,
    datecolumns: list[int],
    locale: str = "en_GB",
) -> tuple[dict, dict]:
    """
    A function that processes a row of a table.

    This is the main worker function of :mod:`basic.csv` module, but is often
    re-used by any other parser that needs to process tabular data.

    Parameters
    ----------
    headers
        A list of headers of the table.

    items
        A list of values corresponding to the headers. Must be the same length as
        headers.

    datefunc
        A function that will generate ``uts`` given a list of values.

    datecolumns
        Column indices that need to be passed to ``datefunc`` to generate uts.

    Returns
    -------
    vals, devs
        A tuple of result dictionaries, with the first element containing the values
        and the second element containing the uncertainties of the values.

    """
    if len(headers) != len(items):
        raise RuntimeError(
            f"Length mismatch between provided headers {headers!r} and items {items!r}."
        )

    vals = {}
    devs = {}
    columns = [column.strip() for column in items]

    # Process raw data, assign sigma and units
    vals["uts"] = datefunc(*[columns[i] for i in datecolumns])
    for ci, header in enumerate(headers):
        if ci in datecolumns:
            continue
        elif columns[ci] == "":
            continue
        try:
            val, dev = tuple_fromstr(str(parse_decimal(columns[ci], locale=locale)))
            vals[header] = val
            devs[header] = dev
        except ValueError:
            vals[header] = columns[ci]

    return vals, devs


def extract(
    *,
    fn: str,
    encoding: str,
    locale: str,
    timezone: str,
    parameters: BaseModel,
    **kwargs: dict,
) -> DataTree:
    if hasattr(parameters, "strip"):
        strip = parameters.strip
    else:
        strip = None

    # Load file, extract headers and get timestamping function
    with open(fn, "r", encoding=encoding) as infile:
        # This decode/encode is done to account for some csv files that have a BOM
        # at the beginning of each line.
        lines = [i.encode().decode(encoding) for i in infile.readlines()]
    assert len(lines) >= 2
    headers = [h.strip().strip(strip) for h in lines[0].split(parameters.sep)]

    datecolumns, datefunc, fulldate = dgutils.infer_timestamp_from(
        headers=headers, spec=parameters.timestamp, timezone=timezone
    )

    # Populate units
    units = parameters.units
    if units is None:
        units = {}
        _units = [c.strip().strip(strip) for c in lines[1].split(parameters.sep)]
        for header in headers:
            units[header] = _units.pop(0)
        si = 2
    else:
        for header in headers:
            if header not in units:
                logger.warning(f"Using implicit dimensionless unit ' ' for {header!r}")
                units[header] = " "
            elif units[header] == "":
                units[header] = " "
        si = 1

    units = dgutils.sanitize_units(units)

    # Process rows
    data_vals = {}
    meta_vals = {"_fn": []}
    for li, line in enumerate(lines[si:]):
        vals, devs = process_row(
            headers,
            [i.strip().strip(strip) for i in line.split(parameters.sep)],
            datefunc,
            datecolumns,
            locale=locale,
        )
        dgutils.append_dicts(vals, devs, data_vals, meta_vals, fn, li)

    return DataTree(dgutils.dicts_to_dataset(data_vals, meta_vals, units, fulldate))



import pytest
import os
import pickle
from yadg.extractors.drycal.csv import extract
from .utils import datagram_from_file, compare_datatrees


@pytest.mark.parametrize(
    "infile",
    [
        "20211011_DryCal_out.csv",
        "20220721-porosity-study-20p-Cu-200mA-EDLC-01-flow.csv",
        "20220912_Defender.csv",
    ],
)
def test_drycal_csv(infile, datadir):
    os.chdir(datadir)
    ret = extract(fn=infile, encoding="utf8", timezone="Europe/Berlin")
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)


def test_drycal_lock_stock_dataschema(datadir):
    os.chdir(datadir)
    ret = datagram_from_file("lock_stock_dataschema.yml")
    print(f"{ret=}")
    assert ret["outlet"]["DryCal"].shape == (187,)



column_units = {
    '"Ri"/Ohm': ("'Ri'", "Ω"),
    "Im(C)/nF": ("Im(C)", "nF"),
    "Im(Conductivity)/mS/cm": ("Im(Conductivity)", "mS/cm"),
    "Im(M)": ("Im(M)", ""),
    "Im(Permittivity)": ("Im(Permittivity)", ""),
    "Im(Resistivity)/Ohm.cm": ("Im(Resistivity)", "Ω·cm"),
    "-Im(Z)/Ohm": ("-Im(Z)", "Ω"),
    "-Im(Zce)/Ohm": ("-Im(Zce)", "Ω"),
    "-Im(Zwe-ce)/Ohm": ("-Im(Zwe-ce)", "Ω"),
    "(Q-Qo)/C": ("(Q-Qo)", "C"),
    "(Q-Qo)/mA.h": ("(Q-Qo)", "mA·h"),
    "<Ece>/V": ("<Ece>", "V"),
    "<Ewe>/V": ("<Ewe>", "V"),
    "<Ewe/V>": ("<Ewe>", "V"),
    "<I>/mA": ("<I>", "mA"),
    "|C|/nF": ("|C|", "nF"),
    "|Conductivity|/mS/cm": ("|Conductivity|", "mS/cm"),
    "|Ece|/V": ("|Ece|", "V"),
    "|Energy|/W.h": ("|Energy|", "W·h"),
    "|Ewe|/V": ("|Ewe|", "V"),
    "|Ewe h2|/V": ("|Ewe h2|", "V"),
    "|Ewe h3|/V": ("|Ewe h3|", "V"),
    "|Ewe h4|/V": ("|Ewe h4|", "V"),
    "|Ewe h5|/V": ("|Ewe h5|", "V"),
    "|Ewe h6|/V": ("|Ewe h6|", "V"),
    "|Ewe h7|/V": ("|Ewe h7|", "V"),
    "|I|/A": ("|I|", "A"),
    "|I h2|/A": ("|I h2|", "A"),
    "|I h3|/A": ("|I h3|", "A"),
    "|I h4|/A": ("|I h4|", "A"),
    "|I h5|/A": ("|I h5|", "A"),
    "|I h6|/A": ("|I h6|", "A"),
    "|I h7|/A": ("|I h7|", "A"),
    "|M|": ("|M|", ""),
    "|Permittivity|": ("|Permittivity|", ""),
    "|Resistivity|/Ohm.cm": ("|Resistivity|", "Ω·cm"),
    "|Y|/Ohm-1": ("|Y|", "S"),
    "|Z|/Ohm": ("|Z|", "Ω"),
    "|Zce|/Ohm": ("|Zce|", "Ω"),
    "|Zwe-ce|/Ohm": ("|Zwe-ce|", "Ω"),
    "Analog IN 1/V": ("Analog IN 1", "V"),
    "Analog IN 2/V": ("Analog IN 2", "V"),
    "Analog OUT/V": ("Analog OUT", "V"),
    "Capacitance charge/µF": ("Capacitance charge", "µF"),
    "Capacitance discharge/µF": ("Capacitance discharge", "µF"),
    "Capacity/mA.h": ("Capacity", "mA·h"),
    "charge time/s": ("charge time", "s"),
    "Conductivity/S.cm-1": ("Conductivity", "S/cm"),
    "control changes": ("control changes", None),
    "control/mA": ("control_I", "mA"),
    "control/V": ("control_V", "V"),
    "control/V/mA": ("control_VI", "mA"),
    "counter inc.": ("counter inc.", None),
    "Cp-2/µF-2": ("Cp⁻²", "µF⁻²"),
    "Cp/µF": ("Cp", "µF"),
    "Cs-2/µF-2": ("Cs⁻²", "µF⁻²"),
    "Cs/µF": ("Cs", "µF"),
    "cycle number": ("cycle number", None),
    "cycle time/s": ("cycle time", "s"),
    "d(Q-Qo)/dE/mA.h/V": ("d(Q-Qo)/dE", "mA·h/V"),
    "dI/dt/mA/s": ("dI/dt", "mA/s"),
    "discharge time/s": ("discharge time", "s"),
    "dQ/C": ("dQ", "C"),
    "dq/mA.h": ("dq", "mA·h"),
    "dQ/mA.h": ("dQ", "mA·h"),
    "Ece/V": ("Ece", "V"),
    "Ecell/V": ("Ecell", "V"),
    "Efficiency/%": ("Efficiency", "%"),
    "Energy charge/W.h": ("Energy charge", "W·h"),
    "Energy discharge/W.h": ("Energy discharge", "W·h"),
    "Energy/W.h": ("Energy", "W·h"),
    "error": ("error", None),
    "Ewe-Ece/V": ("Ewe-Ece", "V"),
    "Ewe/V": ("Ewe", "V"),
    "freq/Hz": ("freq", "Hz"),
    "half cycle": ("half cycle", None),
    "I Range": ("I Range", None),
    "I/mA": ("I", "mA"),
    "Im(Y)/Ohm-1": ("Im(Y)", "S"),
    "Loss Angle(Delta)/deg": ("Loss Angle(Delta)", "deg"),
    "mode": ("mode", None),
    "Ns changes": ("Ns changes", None),
    "Ns": ("Ns", None),
    "NSD Ewe/%": ("NSD Ewe", "%"),
    "NSD I/%": ("NSD I", "%"),
    "NSR Ewe/%": ("NSR Ewe", "%"),
    "NSR I/%": ("NSR I", "%"),
    "ox/red": ("ox or red", None),
    "P/W": ("P", "W"),
    "Phase(C)/deg": ("Phase(C)", "deg"),
    "Phase(Conductivity)/deg": ("Phase(Conductivity)", "deg"),
    "Phase(M)/deg": ("Phase(M)", "deg"),
    "Phase(Permittivity)/deg": ("Phase(Permittivity)", "deg"),
    "Phase(Resistivity)/deg": ("Phase(Resistivity)", "deg"),
    "Phase(Y)/deg": ("Phase(Y)", "deg"),
    "Phase(Z)/deg": ("Phase(Z)", "deg"),
    "Phase(Zce)/deg": ("Phase(Zce)", "deg"),
    "Phase(Zwe-ce)/deg": ("Phase(Zwe-ce)", "deg"),
    "Q charge/discharge/mA.h": ("Q charge or discharge", "mA·h"),
    "Q charge/mA.h": ("Q charge", "mA·h"),
    "Q charge/mA.h/g": ("Q charge", "mA·h/g"),
    "Q discharge/mA.h": ("Q discharge", "mA·h"),
    "Q discharge/mA.h/g": ("Q discharge", "mA·h/g"),
    "R/Ohm": ("R", "Ω"),
    "Rcmp/Ohm": ("Rcmp", "Ω"),
    "Re(C)/nF": ("Re(C)", "nF"),
    "Re(Conductivity)/mS/cm": ("Re(Conductivity)", "mS/cm"),
    "Re(M)": ("Re(M)", ""),
    "Re(Permittivity)": ("Re(Permittivity)", ""),
    "Re(Resistivity)/Ohm.cm": ("Re(Resistivity)", "Ω·cm"),
    "Re(Y)/Ohm-1": ("Re(Y)", "S"),
    "Re(Z)/Ohm": ("Re(Z)", "Ω"),
    "Re(Zce)/Ohm": ("Re(Zce)", "Ω"),
    "Re(Zwe-ce)/Ohm": ("Re(Zwe-ce)", "Ω"),
    "step time/s": ("step time", "s"),
    "Tan(Delta)": ("Tan(Delta)", ""),
    "THD Ewe/%": ("THD Ewe", "%"),
    "THD I/%": ("THD I", "%"),
    "time/s": ("time", "s"),
    "x": ("x", None),
    "z cycle": ("z cycle", None),
}



import os
import pickle
import json
import datetime
import dateutil.parser
import logging
from zoneinfo import ZoneInfo
import numpy as np
from pydantic import BaseModel
from typing import Callable, Union, Mapping, Iterable
from xarray import Dataset
from dgbowl_schemas.yadg.dataschema_5_0.externaldate import ExternalDate
from dgbowl_schemas.yadg.dataschema_5_0.timestamp import TimestampSpec


logger = logging.getLogger(__name__)


def now(
    asstr: bool = False, tz: datetime.timezone = datetime.timezone.utc
) -> Union[float, str]:
    """
    Wrapper around datetime.now()

    A convenience function for returning the current time as a ISO 8601 or as a Unix
    timestamp.
    """
    dt = datetime.datetime.now(tz=tz)
    if asstr:
        return dt.strftime("%Y-%m-%d %H:%M:%S")
    else:
        return dt.timestamp()


def ole_to_uts(ole_timestamp: float, timezone: str) -> float:
    """
    Converts a Microsoft OLE timestamp into a POSIX timestamp.

    The OLE automation date format is a floating point value, counting
    days since midnight 30 December 1899. Hours and minutes are
    represented as fractional days.

    https://devblogs.microsoft.com/oldnewthing/20030905-02/?p=42653

    Parameters
    ----------
    ole_timestamp
        A timestamp in Microsoft OLE format.
    timezone
        String desribing the timezone.

    Returns
    -------
    time: float
        The corresponding Unix timestamp.

    """
    tzinfo = ZoneInfo(timezone)
    ole_base = datetime.datetime(year=1899, month=12, day=30, tzinfo=tzinfo)
    ole_delta = datetime.timedelta(days=ole_timestamp)
    time = ole_base + ole_delta
    return time.timestamp()


def str_to_uts(
    *,
    timestamp: str,
    timezone: str,
    format: str = None,
    strict: bool = True,
) -> Union[float, None]:
    """
    Converts a string to POSIX timestamp.

    If the optional ``format`` is specified, the ``timestamp`` string is processed
    using the :func:`datetime.datetime.strptime` function; if no ``format`` is
    supplied, an ISO 8601 format is assumed and an attempt to parse using
    :func:`dateutil.parser.parse` is made.

    Parameters
    ----------
    timestamp
        A string containing the timestamp.

    format
        Optional format string for parsing of the ``timestamp``.

    timezone
        Optional timezone of the ``timestamp``. By default, "UTC".

    strict
        Whether to re-raise any parsing errors.

    Returns
    -------
    uts: Union[float, None]
        Returns the POSIX timestamp if successful, otherwise None.

    """

    try:
        if format is None:
            dt = dateutil.parser.parse(timestamp)
        else:
            dt = datetime.datetime.strptime(timestamp, format)

        local_tz = ZoneInfo(timezone) if dt.tzinfo is None else dt.tzinfo
        local_dt = dt.replace(tzinfo=local_tz)
        utc_dt = local_dt.astimezone(datetime.timezone.utc)
        return utc_dt.timestamp()
    except (dateutil.parser.ParserError, ValueError) as e:
        if strict:
            raise e
        else:
            logger.info(
                "Parsing timestamp '%s' with format '%s' was not successful.",
                str(timestamp),
                format,
            )
            return None


def infer_timestamp_from(
    *,
    headers: list = None,
    spec: TimestampSpec = None,
    timezone: str,
) -> tuple[list, Callable, bool]:
    """
    Convenience function for timestamping

    Given a set of headers, and an optional specification, return an array containing
    column indices from which a timestamp in a given row can be computed, as well as the
    function which will compute the timestamp given the returned array.

    Parameters
    ----------
    headers
        An array of strings. If `spec` is not supplied, must contain either "uts"
        :class:`(float)` or "timestep" :class:`(str)` (conforming to ISO 8601).

    spec
        A specification of timestamp elements with associated column indices and
        optional formats. Currently accepted combinations of keys are: "uts";
        "timestamp"; "date" and / or "time".

    tz
        Timezone to use for conversion. By default, UTC is used.

    Returns
    -------
    (datecolumns, datefunc, fulldate): tuple[list, Callable, bool]
        A tuple containing a list of indices of columns, a Callable to which the
        columns have to be passed to obtain a uts timestamp, and whether the determined
        timestamp is full or partial.

    """

    if spec is not None:
        if hasattr(spec, "uts"):
            return [spec.uts.index], float, True
        elif hasattr(spec, "timestamp"):

            def retfunc(value):
                return str_to_uts(
                    timestamp=value, format=spec.timestamp.format, timezone=timezone
                )

            return [spec.timestamp.index], retfunc, True
        elif hasattr(spec, "date") or hasattr(spec, "time"):
            specdict = {
                "date": datetime.datetime.fromtimestamp(0).timestamp,
                "time": datetime.datetime.fromtimestamp(0).timestamp,
            }
            cols = [None, None]
            if spec.date is not None:

                def datefn(value):
                    return str_to_uts(
                        timestamp=value, format=spec.date.format, timezone=timezone
                    )

                cols[0] = spec.date.index
                specdict["date"] = datefn
            if spec.time is not None:
                if spec.time.format is not None:

                    def timefn(value):
                        t = datetime.datetime.strptime(value, spec.time.format)
                        td = datetime.timedelta(
                            hours=t.hour,
                            minutes=t.minute,
                            seconds=t.second,
                            microseconds=t.microsecond,
                        )
                        return td.total_seconds()

                else:
                    logger.debug(
                        "Assuming specified column is time in ISO 8601 format."
                    )

                    def timefn(value):
                        t = datetime.time.fromisoformat(value)
                        td = datetime.timedelta(
                            hours=t.hour,
                            minutes=t.minute,
                            seconds=t.second,
                            microseconds=t.microsecond,
                        )
                        return td.total_seconds()

                cols[1] = spec.time.index
                specdict["time"] = timefn
            if cols[0] is None:
                return [cols[1]], specdict["time"], False
            elif cols[1] is None:
                return [cols[0]], specdict["date"], False
            else:

                def retfn(date, time):
                    return specdict["date"](date) + specdict["time"](time)

                return cols, retfn, True
    elif "uts" in headers:
        logger.debug("Assuming column 'uts' is a valid unix timestamp.")
        return [headers.index("uts")], float, True
    elif "timestamp" in headers:
        logger.debug("Assuming column 'timestamp' is a valid ISO 8601 timestamp")

        def retfunc(value):
            return str_to_uts(timestamp=value, timezone=timezone)

        return [headers.index("timestamp")], retfunc, True
    else:
        assert False, "dateutils: A valid timestamp could not be deduced."


def complete_timestamps(
    *,
    timesteps: list,
    fn: str,
    spec: ExternalDate,
    timezone: str,
) -> list[float]:
    """
    Timestamp completing function.

    This function allows for completing or overriding the uts timestamps determined by
    the individual parsers. **yadg** enters this function for any parser which does not
    return a full timestamp, as well as if the ``externaldate`` specification is
    specified by the user.

    The ``externaldate`` specification is as follows:

    .. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.externaldate.ExternalDate
       :noindex:

    The ``using`` key specifies how an external timestamp is created. Only one entry in
    ``using`` is permitted. By default, this entry is:

    .. code-block:: yaml

        using:
          filename:
            format: "%Y-%m-%d-%H-%M-%S"
            len: 19

    Which means the code will attempt to deduce the timestamp from the path of the
    processed file (``fn``), using the first 19 characters of the base filename
    according to the above format (eg. "2021-12-31-13-45-00").

    If ``file`` is specified, the handling of timestamps is handed off to
    :func:`timestamps_from_file`.

    The ``mode`` key specifies whether the offsets determined in this function are
    added to the current timestamps (eg. date offset being added to time) or whether
    they should replace the existing timestamps completely.

    As a measure of last resort, the ``mtime`` of the ``fn`` is used. ``mtime`` is
    preferred to ``ctime``, as the former has a more consistent cross-platform
    behaviour.

    Parameters
    ----------
    timesteps
        A list of timesteps generated from a single file, ``fn``.

    fn
        Filename used to create ``timesteps``.

    spec
        ``externaldate`` specification part of the `schema`.

    timezone
        Timezone, defaults to "UTC".

    """
    delta = None

    fulldate = True
    if spec is not None:
        replace = spec.mode == "replace"
        method = spec.using
    else:
        replace = False
        method = None

    if hasattr(method, "file"):
        delta = timestamps_from_file(
            method.file.path, method.file.type, method.file.match, timezone
        )
    elif hasattr(method, "isostring"):
        delta = str_to_uts(
            timestamp=method.isostring, format=None, timezone=timezone, strict=True
        )
    elif hasattr(method, "utsoffset"):
        delta = method.utsoffset
    elif hasattr(method, "filename"):
        basename = os.path.basename(fn)
        filename, ext = os.path.splitext(basename)
        string = filename[: method.filename.len]
        delta = str_to_uts(
            timestamp=string,
            format=method.filename.format,
            timezone=timezone,
            strict=False,
        )

    if delta is None:
        logger.warning("Timestamp completion failed. Using 'mtime'.")
        delta = os.path.getmtime(fn)
        fulldate = False

    if isinstance(delta, float):
        delta = [delta] * len(timesteps)

    assert len(delta) == len(timesteps)

    if replace:
        timesteps = delta
    else:
        timesteps = timesteps + delta

    return timesteps, fulldate


def timestamps_from_file(
    path: str,
    type: str,
    match: str,
    timezone: str,
) -> Union[float, list[float]]:
    """
    Load timestamps from file.

    This function enables loading timestamps from file specified by the ``path``.
    The currently supported file formats include ``json`` and ``pkl``, which must
    contain a top-level :class:`Mapping` with a key that is matched by ``match``,
    or a top-level :class:`Iterable`, both containing :class:`str` or :class:`float`
    -like objects that can be processed into an Unix timestamp.

    Parameters
    ----------
    path
        Location of the external file.

    type
        Type of the external file. Currently, ``"json", "pkl"`` are supported.

    match
        An optional key to match if the object in ``path`` is a :class:`Mapping`.

    timezone
        An optional timezone string, defaults to "UTC"

    Returns
    -------
    parseddata: Union[float, list[float]]
        A single or a list of POSIX timestamps.

    """
    assert os.path.exists(path), f"timestamps_from_file: path '{path}' doesn't exist."
    assert os.path.isfile(path), f"timestamps_from_file: Path '{path}' is not a file."

    if type == "pkl":
        logger.debug("Loading '%s' as pickle.", path)
        with open(path, "rb") as infile:
            data = pickle.load(infile)
    elif type == "json":
        logger.debug("Loading '%s' as json.", path)
        with open(path, "r") as infile:
            data = json.load(infile)
    elif type == "agilent.log":
        logger.critical("Type 'agilent.log' not yet supported.")
        data = None

    if data is None:
        return data
    elif isinstance(data, Iterable):
        if isinstance(data, Mapping):
            assert match in data
            data = data[match]
        if isinstance(data, Iterable):
            parseddata = []
            for i in data:
                if isinstance(i, str):
                    delta = str_to_uts(
                        timestamp=i, format=None, timezone=timezone, strict=True
                    )
                    parseddata.append(delta)
                elif isinstance(i, (int, float, np.number)):
                    parseddata.append(float(i))
            return parseddata
        else:
            if isinstance(data, str):
                return str_to_uts(
                    timestamp=data, format=None, timezone=timezone, strict=True
                )
            else:
                return float(data)


def complete_uts(
    ds: Dataset,
    filename: str,
    externaldate: BaseModel,
    timezone: str,
) -> Dataset:
    """
    A helper function ensuring that the Dataset ``ds`` contains a dimension ``"uts"``,
    and that the timestamps in ``"uts"`` are completed as instructed in the
    ``externaldate`` specification.

    """
    if not hasattr(ds, "uts"):
        ds = ds.expand_dims("uts")
    if len(ds.uts.coords) == 0:
        ds["uts"] = np.zeros(ds.uts.size)
        ds.attrs["fulldate"] = 0

    # fulldate should be an int, as it's converted int yadg.extract.extract_from_path()
    if ds.attrs.get("fulldate", 1) == 0 or externaldate is not None:
        ts, fulldate = complete_timestamps(
            timesteps=ds.uts.values,
            fn=filename,
            spec=externaldate,
            timezone=timezone,
        )
        ds["uts"] = ts
        if fulldate:
            ds.attrs.pop("fulldate", None)
    return ds



"""
Extractor of Agilent Chemstation Chromtab tabulated data files. This file format may
include multiple timesteps consisting of several traces each in a single CSV file. It
contains a header section for each timestep, followed by a detector name, and a sequence
of "X, Y" datapoints, which are stored as ``elution_time`` and ``signal``.

.. warning ::

    It is not guaranteed that the X-axis of the chromatogram (i.e. ``elution_time``) is
    consistent between the timesteps of the same trace. The traces are expanded to the
    length of the longest trace, and the shorter traces are padded with ``NaNs``.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Agilent_csv

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      {{ detector_name }}:
        coords:
          uts:            !!float               # Unix timestamp
          elution_time:   !!float               # Elution time
        data_vars:
          signal:         (uts, elution_time)   # Signal data

Metadata
````````
The following metadata is extracted:

    - ``sampleid``: Sample name.
    - ``datafile``: Original path of the data file.

Uncertainties
`````````````
All uncertainties are derived from the string representation of the floats.

.. codeauthor::
    Peter Kraus

"""

import numpy as np
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
from yadg import dgutils
import xarray as xr
from datatree import DataTree
import logging

logger = logging.getLogger(__name__)


def _process_headers(headers: list, columns: list, timezone: str) -> dict:
    orig_meta = {k: v.strip() for k, v in zip(headers, columns)}
    if "Date Acquired" in orig_meta:
        uts = dgutils.str_to_uts(
            timestamp=columns[headers.index("Date Acquired")].strip(),
            format="%d %b %Y %H:%M",
            timezone=timezone,
        )
    else:
        logger.warning("'Date Acquired' not in file header, cannot infer timestamp.")
        uts = None
    return orig_meta, uts


def _to_trace(tx, ty):
    tvals, tders = [x for x in zip(*tx)]
    yvals, yders = [x for x in zip(*ty)]
    trace = {
        "tvals": np.array(tvals) * 60,
        "tdevs": np.array(tders) * 60,
        "yvals": list(yvals),
        "ydevs": list(yders),
    }
    return trace


def extract(
    *,
    fn: str,
    encoding: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()
    orig_meta = {}
    uts = []
    tx = []
    ty = []
    detname = None
    tstep = dict()
    data = []
    traces = set()
    maxlen = dict()
    for line in lines:
        parts = line.strip().split(",")
        if len(parts) > 2:
            if '"Date Acquired"' in parts:
                if tx != [] and ty != [] and detname is not None:
                    trace = _to_trace(tx, ty)
                    tstep[detname] = trace
                    maxlen[detname] = max(maxlen.get(detname, 0), len(trace["tvals"]))
                    tx = []
                    ty = []
                if len(tstep) > 0:
                    data.append(tstep)
                    tstep = dict()
                headers = [p.replace('"', "") for p in parts]
            else:
                columns = [p.replace('"', "") for p in parts]
                ret = _process_headers(headers, columns, timezone)
                uts.append(ret[1])
                dgutils.merge_meta(orig_meta, ret[0])
        elif len(parts) == 1:
            if tx != [] and ty != [] and detname is not None:
                trace = _to_trace(tx, ty)
                tstep[detname] = trace
                maxlen[detname] = max(maxlen.get(detname, 0), len(trace["tvals"]))
                tx = []
                ty = []
            detname = parts[0].replace('"', "").split("\\")[-1]
            traces.add(detname)
        elif len(parts) == 2:
            x, y = [tuple_fromstr(i) for i in parts]
            tx.append(x)
            ty.append(y)
    trace = _to_trace(tx, ty)
    tstep[detname] = trace
    maxlen[detname] = max(maxlen.get(detname, 0), len(trace["tvals"]))
    data.append(tstep)

    traces = sorted(traces)
    vals = {}
    for tr in traces:
        dsets = []
        for ti, ts in enumerate(data):
            thislen = len(ts[tr]["tvals"])
            fvals = {}
            for k in {"yvals", "ydevs", "tvals", "tdevs"}:
                fvals[k] = np.ones(maxlen[tr]) * np.nan
                fvals[k][:thislen] = ts[tr][k]
            ds = xr.Dataset(
                data_vars={
                    "signal": (
                        ["elution_time"],
                        fvals["yvals"],
                        {"ancillary_variables": "signal_std_err"},
                    ),
                    "signal_std_err": (
                        ["elution_time"],
                        fvals["ydevs"],
                        {"standard_name": "signal standard_error"},
                    ),
                    "elution_time": (
                        ["_"],
                        fvals["tvals"],
                        {"units": "s", "ancillary_variables": "elution_time_std_err"},
                    ),
                    "elution_time_std_err": (
                        ["elution_time"],
                        fvals["tdevs"],
                        {"units": "s", "standard_name": "elution_time standard_error"},
                    ),
                },
                coords={},
                attrs={},
            )
            ds["uts"] = [uts[ti]]
            dsets.append(ds)
        vals[tr] = xr.concat(dsets, dim="uts")
    dt = DataTree.from_dict(vals)
    dt.attrs = {"original_metadata": orig_meta}
    return dt



from .helpers import get_yadg_metadata
from .dateutils import (
    now,
    infer_timestamp_from,
    str_to_uts,
    ole_to_uts,
    complete_timestamps,
    complete_uts,
)
from .schemautils import update_schema, schema_from_preset
from .btools import read_value
from .pintutils import sanitize_units, ureg
from .dsutils import dicts_to_dataset, append_dicts, merge_dicttrees, merge_meta

__all__ = [
    "get_yadg_metadata",
    "now",
    "infer_timestamp_from",
    "str_to_uts",
    "ole_to_uts",
    "complete_timestamps",
    "complete_uts",
    "update_schema",
    "schema_from_preset",
    "read_value",
    "sanitize_units",
    "ureg",
    "dicts_to_dataset",
    "append_dicts",
    "merge_dicttrees",
    "merge_meta",
]



"""
This parser handles the reading and processing of the legacy log files created by
the LabView interface for the MCPT instrument at FHI, now FU Berlin. These files contain
information about the timestamp, temperatures, and inlet / process flows.

Usage
`````
Available since ``yadg-3.0``. Deprecated since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.FHI_csv

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float     # Unix timestamp
      data_vars:
        T_f:            (uts)       # Flow temperature
        T_fs:           (uts)       # Flow temperature setpoint
        T_fo:           (uts)       # Flow heater duty cycle
        T_c:            (uts)       # Cavity temperature
        T_cs:           (uts)       # Cavity temperature setpoint
        T_co:           (uts)       # Cavity cooling duty cycle
        T_cal:          (uts)       # Calibration thermocouple temperature
        N2:             (uts)       # N2 flow
        O2:             (uts)       # N2 flow
        alkane:         (uts)       # alkane flow
        CO_CO2:         (uts)       # CO or CO2 flow
        saturator:      (uts)       # saturator flow
        pressure:       (uts)       # Reactor flow meter back-pressure
        flow low:       (uts)       # Reactor mix high-flow MFC
        flow high:      (uts)       # Reactor mix low-flow MFC
        cavity flush:   (uts)       # Cavity N2 flow
        heater flow:    (uts)       # Heater flow

Metadata
````````
No metadata is returned.

.. codeauthor::
    Peter Kraus

"""

import logging
from pydantic import BaseModel
from yadg.extractors.basic.csv import process_row
from yadg import dgutils
from datatree import DataTree

logger = logging.getLogger(__name__)


def extract(
    *,
    fn: str,
    encoding: str,
    timezone: str,
    parameters: BaseModel,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r", encoding=encoding) as infile:
        lines = [i.strip() for i in infile.readlines()]

    headers = [i.strip() for i in lines.pop(0).split(";")]

    for hi, header in enumerate(headers):
        if "/" in header:
            logger.warning("Replacing '/' for '_' in header '%s'.", header)
            headers[hi] = header.replace("/", "_")

    _units = [i.strip() for i in lines.pop(0).split(";")]
    units = {}
    for h in headers:
        units[h] = _units.pop(0)

    units = dgutils.sanitize_units(units)

    datecolumns, datefunc, fulldate = dgutils.infer_timestamp_from(
        spec=parameters.timestamp,
        timezone=timezone,
    )

    # Process rows
    data_vals = {}
    meta_vals = {"_fn": []}
    for li, line in enumerate(lines):
        vals, devs = process_row(
            headers,
            line.split(";"),
            datefunc,
            datecolumns,
        )
        dgutils.append_dicts(vals, devs, data_vals, meta_vals, fn, li)

    return DataTree(dgutils.dicts_to_dataset(data_vals, meta_vals, units, fulldate))



"""
For processing of BioLogic's EC-Lab binary modular files.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.EClab_mpr

Schema
``````
The ``mpr`` files contain many columns that vary depending on the electrochemical
technique used. Below is shown a list of columns that can be expected to be present
in a typical ``mpr`` file.

.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float     # Unix timestamp, without date
      data_vars:
        Ewe             (uts)       # Potential of the working electrode
        Ece             (uts)       # Potential of the counter electrode, if present
        I               (uts)       # Instantaneous current
        time            (uts)       # Time elapsed since the start of the experiment
        <Ewe>           (uts)       # Average Ewe potential since last data point
        <Ece>           (uts)       # Average Ece potential since last data point
        <I>             (uts)       # Average current since last data point
        ...

.. note::

     Note that in most cases, either the instantaneous or the averaged quantities are
     stored - only rarely are both available!

Notes on file structure
```````````````````````
``.mpr`` files are structured in a set of "modules", one concerning
settings, one for actual data, one for logs, and an optional loops
module. The parameter sequences can be found in the settings module.

This code is partly an adaptation of the `galvani module by Chris
Kerr <https://github.com/echemdata/galvani>`_, and builds on the work done
by the previous civilian service member working on the project, Jonas Krieger.

These are the implemented techniques for which the technique parameter
sequences can be parsed:

+------+-------------------------------------------------+
| CA   | Chronoamperometry / Chronocoulometry            |
+------+-------------------------------------------------+
| CP   | Chronopotentiometry                             |
+------+-------------------------------------------------+
| CV   | Cyclic Voltammetry                              |
+------+-------------------------------------------------+
| GCPL | Galvanostatic Cycling with Potential Limitation |
+------+-------------------------------------------------+
| GEIS | Galvano Electrochemical Impedance Spectroscopy  |
+------+-------------------------------------------------+
| LOOP | Loop                                            |
+------+-------------------------------------------------+
| LSV  | Linear Sweep Voltammetry                        |
+------+-------------------------------------------------+
| MB   | Modulo Bat                                      |
+------+-------------------------------------------------+
| OCV  | Open Circuit Voltage                            |
+------+-------------------------------------------------+
| PEIS | Potentio Electrochemical Impedance Spectroscopy |
+------+-------------------------------------------------+
| WAIT | Wait                                            |
+------+-------------------------------------------------+
| ZIR  | IR compensation (PEIS)                          |
+------+-------------------------------------------------+

At a top level, ``.mpr`` files are made up of a number of modules,
separated by the ``MODULE`` keyword. In all the files I have seen, the
first module is the settings module, followed by the data module, the
log module and then an optional loop module.

.. code-block::

    0x0000 BIO-LOGIC MODULAR FILE  # File magic.
    0x0034 MODULE                  # Module magic.
    ...                            # Module 1.
    0x???? MODULE                  # Module magic.
    ...                            # Module 2.
    0x???? MODULE                  # Module magic.
    ...                            # Module 3.
    0x???? MODULE                  # Module magic.
    ...                            # Module 4.

After splitting the entire file on ``MODULE``, each module starts with a
header that is structured like this (offsets from start of module):

.. code-block::

    0x0000 short_name  # Short name, e.g. VMP Set.
    0x000A long_name   # Longer name, e.g. VMP settings.
    0x0023 length      # Number of bytes in module data.
    0x0027 version     # Module version.
    0x002B date        # Acquisition date in ASCII, e.g. 08/10/21.
    ...                # Module data.

The contents of each module's data vary wildly depending on the used
technique, the module and perhaps the software version, the settings in
EC-Lab, etc. Here a quick overview (offsets from start of module data).

*Settings Module*

.. code-block::

    0x0000 technique_id           # Unique technique ID.
    ...                           # ???
    0x0007 comments               # Pascal string.
    ...                           # Zero padding.
    # Cell Characteristics.
    0x0107 active_material_mass   # Mass of active material
    0x010B at_x                   # at x =
    0x010F molecular_weight       # Molecular weight of active material
    0x0113 atomic_weight          # Atomic weight of intercalated ion
    0x0117 acquisition_start      # Acquisition started a: xo =
    0x011B e_transferred          # Number of e- transferred
    0x011E electrode_material     # Pascal string.
    ...                           # Zero Padding
    0x01C0 electrolyte            # Pascal string.
    ...                           # Zero Padding, ???.
    0x0211 electrode_area         # Electrode surface area
    0x0215 reference_electrode    # Pascal string
    ...                           # Zero padding
    0x024C characteristic_mass    # Characteristic mass
    ...                           # ???
    0x025C battery_capacity       # Battery capacity C =
    0x0260 battery_capacity_unit  # Unit of the battery capacity.
    ...                           # ???
    # Technique parameters can randomly be found at 0x0572, 0x1845 or
    # 0x1846. All you can do is guess and try until it fits.
    0x1845 ns                     # Number of sequences.
    0x1847 n_params               # Number of technique parameters.
    0x1849 params                 # ns sets of n_params parameters.
    ...                           # ???

*Data Module*

.. code-block::

    0x0000 n_datapoints   # Number of datapoints.
    0x0004 n_columns      # Number of values per datapoint.
    0x0005 column_ids     # n_columns unique column IDs.
    ...
    # Depending on module version, datapoints start 0x195, 0x196, or 0x3ef
    # Length of each datapoint depends on number and IDs of columns.
    0x0195 datapoints     # n_datapoints points of data.

*Log Module*

.. code-block::

    ...                         # ???
    0x0009 channel_number       # Zero-based channel number.
    ...                         # ???
    0x00AB channel_sn           # Channel serial number.
    ...                         # ???
    0x01F8 Ewe_ctrl_min         # Ewe ctrl range min.
    0x01FC Ewe_ctrl_max         # Ewe ctrl range max.
    ...                         # ???
    0x0249 ole_timestamp        # Timestamp in OLE format.
    0x0251 filename             # Pascal String.
    ...                         # Zero padding, ???.
    0x0351 host                 # IP address of host, Pascal string.
    ...                         # Zero padding.
    0x0384 address              # IP address / COM port of potentiostat.
    ...                         # Zero padding.
    0x03B7 ec_lab_version       # EC-Lab version (software)
    ...                         # Zero padding.
    0x03BE server_version       # Internet server version (firmware)
    ...                         # Zero padding.
    0x03C5 interpreter_version  # Command interpretor version (firmware)
    ...                         # Zero padding.
    0x03CF device_sn            # Device serial number.
    ...                         # Zero padding.
    0x0922 averaging_points     # Smooth data on ... points.
    ...                         # ???

*Loop Module*

.. code-block::

    0x0000 n_indexes  # Number of loop indexes.
    0x0004 indexes    # n_indexes indexes at which loops start in data.
    ...               # ???


Metadata
````````
The metadata will contain the information from the *Settings module*. This should
include information about the technique, as well as any explicitly parsed cell
characteristics data specified in EC-Lab.

.. admonition:: TODO

    https://github.com/dgbowl/yadg/issues/12

    The mapping between metadata parameters between ``.mpr`` and ``.mpt`` files
    is not yet complete. In ``.mpr`` files, some technique parameters in the settings
    module correspond to entries in drop-down lists in EC-Lab. These values are
    stored as single-byte values in ``.mpr`` files.

The metadata also contains the infromation from the *Log module*, which contains
more general parameters, like software, firmware and server versions, channel number,
host address and an acquisition start timestamp in Microsoft OLE format.

.. note::

    If the ``.mpr`` file contains an ``ExtDev`` module (containing parameters
    of any external sensors plugged into the device), the ``log`` is usually
    not present and therefore the full timestamp cannot be calculated.

.. codeauthor::
    Nicolas Vetsch

"""

import logging
from datatree import DataTree
import numpy as np
from yadg import dgutils
from .common.techniques import (
    technique_params_dtypes,
    param_from_key,
    get_devs,
)
from .common.mpr_columns import (
    module_header_dtypes,
    settings_dtypes,
    flag_columns,
    data_columns,
    log_dtypes,
    extdev_dtypes,
)

logger = logging.getLogger(__name__)


def process_settings(data: bytes, minver: str) -> tuple[dict, list]:
    """Processes the contents of settings modules.

    Parameters
    ----------
    data
        The data to parse through.

    Returns
    -------
    dict
        The parsed settings.

    """
    settings = {}
    # First parse the settings right at the top of the data block.
    technique, params_dtypes = technique_params_dtypes[data[0x0000]]
    logger.debug("Found technique '%s'.", technique)
    settings["technique"] = technique
    for offset, (dtype, name) in settings_dtypes.items():
        settings[name] = dgutils.read_value(data, offset, dtype)
    # Then determine the technique parameters. The parameters' offset
    # changes depending on the technique present and apparently on some
    # other factor that is unclear to me.
    params_offset = None
    offsets = (
        0x0572,
        0x1845,
        0x1846,
        0x1847,
    )
    for offset in offsets:
        n_params = dgutils.read_value(data, offset + 0x0002, "<u2")
        logger.debug("Trying to find %d technique params at 0x%x.", n_params, offset)
        for dtype, versions in params_dtypes:
            if minver in versions and len(dtype) == n_params:
                params_dtype, params_offset = dtype, offset
                logger.debug("Determined %d parameters at 0x%x.", n_params, offset)
                break
        if params_offset is not None:
            break
    if params_offset is None:
        raise NotImplementedError("Unknown parameter offset or technique dtype.")
    logger.debug("Reading number of parameter sequences at 0x%x.", params_offset)
    ns = dgutils.read_value(data, params_offset, "<u2")
    logger.debug("Reading %d parameter sequences of %d parameters.", ns, n_params)
    rawparams = np.frombuffer(
        data,
        offset=params_offset + 0x0004,
        dtype=params_dtype,
        count=ns,
    )
    pardicts = [dict(zip(value.dtype.names, value.item())) for value in rawparams]
    params = []
    for pardict in pardicts:
        for k, v in pardict.items():
            # MPR quirk: I_range off by one
            if k == "I_range":
                v += 1
            pardict[k] = param_from_key(k, v, to_str=True)
            # Handle NaNs and +/-Inf in params here
            if np.isnan(v) or np.isinf(v):
                pardict[k] = str(v)
        params.append(pardict)
    return settings, params


def parse_columns(column_ids: list[int]) -> tuple[list, list, list, dict]:
    """Puts together column info from a list of data column IDs.

    Note
    ----
    The binary layout of the data in the `.mpr` file is described by a
    sequence of column IDs. Some column IDs relate to (flags) which are
    all packed into a single byte.

    Parameters
    ----------
    column_ids
        A list of column IDs.

    Returns
    -------
    tuple[list, list, list, dict]
        The column names, dtypes, units and a dictionary of flag names
        and bitmasks.

    """
    names = []
    dtypes = []
    units = []
    flags = {}
    for id in column_ids:
        if id in flag_columns:
            bitmask, name = flag_columns[id]
            flags[name] = bitmask
            # Flags column only needs to be added once.
            if "flags" not in names:
                names.append("flags")
                dtypes.append("|u1")
                units.append(None)
        elif id in data_columns:
            dtype, name, unit = data_columns[id]
            if name in names:
                logger.warning("Duplicate column '%s' with unit '%s'.", name, unit)
                name = f"duplicate {name}"
            names.append(name)
            dtypes.append(dtype)
            units.append(unit)
        else:
            name = f"unknown_{len(names)}"
            logger.warning("Unknown column ID %d was assigned into '%s'.", id, name)
            names.append(name)
            dtypes.append("<f4")
            units.append(None)
    return names, dtypes, units, flags


def process_data(
    data: bytes,
    version: int,
    Eranges: list[float],
    Iranges: list[float],
    controls: list[str],
):
    """Processes the contents of data modules.

    Parameters
    ----------
    data
        The data to parse through.

    version
        Module version from the data module header.

    Returns
    -------
    list[dict]
        Processed data ([{column -> value}, ..., {column -> value}]). If
        the column unit is set to None, the value is an int. Otherwise,
        the value is a dict with value ("n"), sigma ("s"), and unit
        ("u").

    """
    n_datapoints = dgutils.read_value(data, 0x0000, "<u4")
    n_columns = dgutils.read_value(data, 0x0004, "|u1")
    if version in {10, 11}:
        column_ids = np.frombuffer(data, offset=0x005, dtype=">u2", count=n_columns)
    elif version in {2, 3}:
        column_ids = np.frombuffer(data, offset=0x005, dtype="<u2", count=n_columns)
    logger.debug("Found %d columns with IDs: %s", n_columns, column_ids)
    # Length of each datapoint depends on number and IDs of columns.
    namelist, dtypelist, unitlist, flaglist = parse_columns(column_ids)
    units = {k: v for k, v in zip(namelist, unitlist) if v is not None}
    data_dtype = np.dtype(list(zip(namelist, dtypelist)))
    # Depending on module version, datapoints start at different offsets.
    if version in {10, 11}:
        offset = 0x3EF
    elif version == 2:
        offset = 0x195
    elif version == 3:
        offset = 0x196
    else:
        raise NotImplementedError(f"Unknown data module version: {version}")
    allvals = dict()
    allmeta = dict()
    values = np.frombuffer(data, offset=offset, dtype=data_dtype, count=n_datapoints)
    values = [dict(zip(value.dtype.names, value.item())) for value in values]
    warn_I_range = False
    warn_Ns = False
    for vi, vals in enumerate(values):
        # Lets split this into two loops: get the indices first, then the data
        for (name, value), unit in list(zip(vals.items(), unitlist)):
            if name.startswith("unknown_"):
                continue
            elif unit is None:
                intv = int(value)
                if name == "I Range":
                    vals[name] = param_from_key("I_range", intv)
                else:
                    vals[name] = intv
        if flaglist:
            flag_bits = vals.pop("flags")
            for name, bitmask in flaglist.items():
                # Two's complement hack to find the position of the
                # rightmost set bit.
                shift = (bitmask & -bitmask).bit_length() - 1
                # Rightshift flag by that amount.
                vals[name] = (flag_bits & bitmask) >> shift

        Ns = vals.get("Ns", 0)
        # Manually merged/appended mpr files have a mysteriously larger Ns
        if Ns >= len(Eranges):
            warn_Ns = True
            Ns = len(Eranges) - 1
        Erange = Eranges[Ns]
        Irstr = Iranges[Ns]
        if "I Range" in vals:
            Irstr = vals["I Range"]
        Irange = param_from_key("I_range", Irstr, to_str=False)

        # I Range can be None if it's set to "Auto", "PAC" or other such string.
        if Irange is None:
            warn_I_range = True
            Irange = 1.0

        if "control_V_I" in vals:
            icv = controls[Ns]
            name = "control_I" if icv in {"I", "C"} else "control_V"  # f"control_{icv}"
            vals[name] = vals.pop("control_V_I")
            units[name] = "mA" if icv in {"I", "C"} else "V"
        devs = get_devs(vals=vals, units=units, Erange=Erange, Irange=Irange)
        dgutils.append_dicts(vals, devs, allvals, allmeta, li=vi)
    if warn_I_range:
        logger.warning("I Range could not be understood, defaulting to 1 A.")
    if warn_Ns:
        logger.warning("Ns found in data exceeds Ns in header, using last defined Ns.")

    ds = dgutils.dicts_to_dataset(allvals, allmeta, units, fulldate=False)
    return ds


def process_log(data: bytes) -> dict:
    """Processes the contents of log modules.

    Parameters
    ----------
    data
        The data to parse through.

    Returns
    -------
    dict
        The parsed log.

    """
    log = {}
    for offset, (dtype, name) in log_dtypes.items():
        log[name] = dgutils.read_value(data, offset, dtype)
    return log


def process_loop(data: bytes) -> dict:
    """Processes the contents of loop modules.

    Parameters
    ----------
    data
        The data to parse through.

    Returns
    -------
    dict
        The parsed loops.

    """
    n_indexes = dgutils.read_value(data, 0x0000, "<u4")
    indexes = np.frombuffer(data, offset=0x0004, dtype="<u4", count=n_indexes)
    return {"n_indexes": n_indexes, "indexes": indexes}


def process_ext(data: bytes) -> dict:
    """Processes the contents of external device modules.

    Parameters
    ----------
    data
        The data to parse through.

    Returns
    -------
    dict
        The parsed log.

    """
    ext = {}
    for offset, (dtype, name) in extdev_dtypes.items():
        ext[name] = dgutils.read_value(data, offset, dtype)

    return ext


def process_modules(contents: bytes) -> tuple[dict, list, list, dict, dict]:
    """Handles the processing of all modules.

    Parameters
    ----------
    contents
        The contents of an .mpr file, minus the file magic.

    Returns
    -------
    tuple[dict, list, dict, dict]
        The processed settings, data, log, and loop modules. If they are
        not present in the provided modules, returns None instead.

    """
    modules = contents.split(b"MODULE")[1:]
    settings = log = loop = ext = None
    for module in modules:
        for mhd in module_header_dtypes:
            header = dgutils.read_value(module, 0x0000, mhd)
            if len(module) == mhd.itemsize + header["length"]:
                version = header.get("newver", 0) + header["oldver"]
                logger.debug(
                    "Parsed module header with length %d, version %s",
                    header["length"],
                    version,
                )
                break
        else:
            raise RuntimeError("Unknown module header.")
        name = header["short_name"].strip()
        # We need to determine file version from the header to be able to select correct
        # dtypes. Unfortunately, the header["oldver"] of the "VMP Set" module is always
        # set to 0. However, the newer versions of this module include the "max_length"
        # entry as well as an "newver" key set to 10.
        if version >= 10:
            minver = "11.50"
        # The oldest version we have in test files is 10.40.
        else:
            minver = "10.40"

        logger.debug("Read '%s' with version '%s' ('%s')", name, version, minver)
        module_data = module[mhd.itemsize :]
        if name == "VMP Set":
            settings, params = process_settings(module_data, minver)
            Eranges = []
            Iranges = []
            ctrls = []
            for el in params:
                E_range_max = el.get("E_range_max", float("inf"))
                E_range_min = el.get("E_range_min", float("-inf"))
                Eranges.append(E_range_max - E_range_min)
                Iranges.append(el.get("I_range", "Auto"))
                if "set_I/C" in el:
                    ctrls.append(el["set_I/C"])
                elif "apply_I/C" in el:
                    ctrls.append(el["apply_I/C"])
                else:
                    ctrls.append(None)
        elif name == "VMP data":
            ds = process_data(module_data, version, Eranges, Iranges, ctrls)
        elif name == "VMP LOG":
            log = process_log(module_data)
        elif name == "VMP loop":
            loop = process_loop(module_data)
        elif name == "VMP ExtDev":
            ext = process_ext(module_data)
        else:
            raise NotImplementedError(f"Unknown module: {name}.")
    if ext is not None:
        settings.update(ext)
    return settings, params, ds, log, loop


def extract(
    *,
    fn: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    file_magic = b"BIO-LOGIC MODULAR FILE\x1a                         \x00\x00\x00\x00"
    with open(fn, "rb") as mpr_file:
        assert mpr_file.read(len(file_magic)) == file_magic, "invalid file magic"
        mpr = mpr_file.read()
    settings, params, ds, log, loop = process_modules(mpr)
    assert settings is not None, "no settings module"
    assert ds is not None, "no data module"
    # Arrange all the data into the correct format.
    # TODO: Metadata could be handled in a nicer way.
    metadata = {"settings": settings, "params": params}
    if log is None:
        logger.warning("No 'log' module present in mpr file. Timestamps incomplete.")
        start_time = 0
        fulldate = False
    else:
        metadata["log"] = log
        start_time = dgutils.ole_to_uts(log["ole_timestamp"], timezone=timezone)
        fulldate = True
    if "time" in ds:
        ds["uts"] = ds["time"] + start_time
    else:
        ds["uts"] = [start_time]
    if fulldate:
        del ds.attrs["fulldate"]
    ds.attrs["original_metadata"] = metadata
    return DataTree(ds)



import pytest
import os
import xarray as xr
import shutil
from yadg.extractors.eclab.mpr import extract as extract_mpr
from yadg.extractors.eclab.mpt import extract as extract_mpt


@pytest.fixture
def _datadir(tmpdir, request):
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)
    base_dir, _ = os.path.split(test_dir)
    for ddir in ["test_x_eclab_mpr", "test_x_eclab_mpt"]:
        shutil.copytree(os.path.join(base_dir, ddir), str(tmpdir), dirs_exist_ok=True)
    return tmpdir


@pytest.mark.parametrize(
    "afile, bfile, locale",
    [
        ("ca.mpr", "ca.mpt", "en_US"),
        ("ca.issue_134.mpr", "ca.issue_134.mpt", "en_US"),
        ("ca.issue_149.mpr", "ca.issue_149.mpt", "de_DE"),
        ("cp.mpr", "cp.mpt", "en_US"),
        ("cp.issue_61.mpr", "cp.issue_61.mpt", "en_US"),
        ("cp.issue_149.mpr", "cp.issue_149.mpt", "de_DE"),
        ("cv.mpr", "cv.mpt", "en_US"),
        ("cv.issue_149.mpr", "cv.issue_149.mpt", "de_DE"),
        ("cva.issue_135.mpr", "cva.issue_135.mpt", "en_US"),
        ("gcpl.mpr", "gcpl.mpt", "en_US"),
        ("gcpl.issue_149.mpr", "gcpl.issue_149.mpt", "de_DE"),
        ("geis.mpr", "geis.mpt", "en_US"),
        ("lsv.mpr", "lsv.mpt", "en_US"),
        ("mb.mpr", "mb.mpt", "en_US"),
        ("mb.issue_95.mpr", "mb.issue_95.mpt", "en_US"),
        ("mb.issue_149.mpr", "mb.issue_149.mpt", "de_DE"),
        ("ocv.mpr", "ocv.mpt", "en_US"),
        ("ocv.issue_149.mpr", "ocv.issue_149.mpt", "de_DE"),
        ("peis.mpr", "peis.mpt", "en_US"),
        ("wait.mpr", "wait.mpt", "en_US"),
        ("zir.mpr", "zir.mpt", "en_US"),
        ("vsp_ocv_wo.mpr", "vsp_ocv_wo.mpt", "en_US"),
        ("vsp_ocv_with.mpr", "vsp_ocv_with.mpt", "en_US"),
        ("vsp_peis_with.mpr", "vsp_peis_with.mpt", "en_US"),
    ],
)
def test_eclab_consistency(afile, bfile, locale, _datadir):
    os.chdir(_datadir)
    aret = extract_mpr(fn=afile, timezone="Europe/Berlin")
    bret = extract_mpt(
        fn=bfile, timezone="Europe/Berlin", encoding="windows-1252", locale=locale
    )
    print(f"{aret.data_vars=}")
    print(f"{bret.data_vars=}")
    for key in aret.variables:
        if key.endswith("std_err"):
            continue
        try:
            xr.testing.assert_allclose(aret[key], bret[key])
        except AssertionError as e:
            e.args = (e.args[0] + f"Error happened on key: {key!r}\n",)
            raise e


@pytest.mark.parametrize(
    "afile, bfile, locale",
    [
        ("geis.issue_149.mpr", "geis.issue_149.mpt", "de_DE"),
        ("peis.issue_149.mpr", "peis.issue_149.mpt", "de_DE"),
    ],
)
def test_eclab_consistency_partial(afile, bfile, locale, _datadir):
    os.chdir(_datadir)
    aret = extract_mpr(fn=afile, timezone="Europe/Berlin")
    bret = extract_mpt(
        fn=bfile, timezone="Europe/Berlin", encoding="windows-1252", locale=locale
    )
    print(f"{aret.data_vars=}")
    print(f"{bret.data_vars=}")
    for key in {
        "freq",
        "Re(Z)",
        "-Im(Z)",
        "|Z|",
        "Phase(Z)",
        "time",
        "<Ewe>",
        "<I>",
        "Cs",
        "Cp",
        "cycle number",
        "|Ewe|",
        "|I|",
        "Ns",
        "I Range",
    }:
        try:
            xr.testing.assert_allclose(aret[key], bret[key])
        except AssertionError as e:
            e.args = (e.args[0] + f"Error happened on key: {key!r}\n",)
            raise e


@pytest.mark.parametrize(
    "afile, bfile, locale",
    [
        ("gcpl.pr_182.1.mpr", "gcpl.pr_182.1.mpt", "en_US"),
        ("gcpl.pr_182.2.mpr", "gcpl.pr_182.2.mpt", "en_US"),
    ],
)
def test_eclab_consistency_partial_2(afile, bfile, locale, _datadir):
    os.chdir(_datadir)
    aret = extract_mpr(fn=afile, timezone="Europe/Berlin")
    bret = extract_mpt(
        fn=bfile, timezone="Europe/Berlin", encoding="windows-1252", locale=locale
    )
    print(f"{aret.data_vars=}")
    print(f"{bret.data_vars=}")
    for key in aret.variables:
        if key.endswith("std_err"):
            continue
        elif key in {"control_I"}:
            continue
        try:
            xr.testing.assert_allclose(aret[key], bret[key])
        except AssertionError as e:
            e.args = (e.args[0] + f"Error happened on key: {key!r}\n",)
            raise e



"""
For processing Inficon Fusion csv export format (csv). This is a tabulated format,
including the concentrations, mole fractions, peak areas, and retention times.

.. warning::

    As also mentioned in the ``csv`` files themselves, the use of this filetype
    is discouraged, and the ``json`` files (or a zipped archive of them) should
    be parsed instead.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Fusion_csv

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:                !!float               # Unix timestamp
        species:            !!str                 # Species name
      data_vars:
        area:               (uts, species)        # Integrated peak area
        concentration:      (uts, species)        # Calibrated peak area
        xout:               (uts, species)        # Mole fraction (normalised conc.)
        retention time:     (uts, species)        # Retention time

Metadata
````````
The following metadata is extracted:

    - ``method``: Name of the chromatographic method.

Uncertainties
`````````````
The uncertainties are derived from the string representation of the floats.

.. codeauthor::
    Peter Kraus

"""

import logging
import numpy as np
import xarray as xr
from datatree import DataTree
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr

from yadg import dgutils


logger = logging.getLogger(__name__)

data_names = {
    "Concentration": "concentration",
    "NormalizedConcentration": "xout",
    "Area": "area",
    "RT(s)": "retention time",
}

data_units = {
    "concentration": "%",
    "xout": "%",
    "area": None,
    "retention time": "s",
}


def extract(
    *,
    fn: str,
    encoding: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()

    data = []
    species = set()
    for line in lines[3:]:
        if "SampleName" in line:
            header = [i.strip() for i in line.split(",")]
            sni = header.index("SampleName")
            method = header[0]
            for ii, i in enumerate(header):
                if i == "":
                    header[ii] = header[ii - 1]
        elif "Detectors" in line:
            detectors = [i.replace('"', "").strip() for i in line.split(",")]
            for ii, i in enumerate(detectors):
                if i == "":
                    detectors[ii] = detectors[ii - 1]
        elif "Time" in line:
            samples = [i.replace('"', "").strip() for i in line.split(",")]
            time = samples[0]
            if time == "Time (GMT 120 mins)":
                offset = "+02:00"
            elif time == "Time (GMT 60 mins)":
                offset = "+01:00"
            else:
                logger.error("offset '%s' not understood", time)
                offset = "+00:00"
        elif "% RSD" in line:
            continue
        else:
            items = line.split(",")
            point = {
                "concentration": {},
                "xout": {},
                "area": {},
                "retention time": {},
                "sampleid": items[sni],
                "uts": dgutils.str_to_uts(
                    timestamp=f"{items[0]}{offset}", timezone=timezone
                ),
            }
            for ii, i in enumerate(items[2:]):
                ii += 2
                species.add(samples[ii])
                point[data_names[header[ii]]][samples[ii]] = tuple_fromstr(i)
            data.append(point)

    species = sorted(species)
    data_vars = {}
    for kk in {"concentration", "xout", "area", "retention time"}:
        vals = []
        devs = []
        for i in range(len(data)):
            ivals, idevs = zip(
                *[data[i][kk].get(cn, (np.nan, np.nan)) for cn in species]
            )
            vals.append(ivals)
            devs.append(idevs)
        data_vars[kk] = (
            ["uts", "species"],
            vals,
            {"anciliary_variables": f"{kk}_std_err"},
        )
        data_vars[f"{kk}_std_err"] = (
            ["uts", "species"],
            devs,
            {"standard_name": f"{kk} standard_error"},
        )
        if data_units[kk] is not None:
            data_vars[kk][2]["units"] = data_units[kk]
            data_vars[f"{kk}_std_err"][2]["units"] = data_units[kk]

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={
            "species": (["species"], species),
            "uts": (["uts"], [i["uts"] for i in data]),
        },
        attrs=dict(original_metadata=dict(method=method)),
    )

    return DataTree(ds)



#!/usr/bin/env python
import argparse
import logging
from importlib import metadata

from yadg import subcommands

logger = logging.getLogger(__name__)


def set_loglevel(delta: int):
    loglevel = min(max(30 - (10 * delta), 10), 50)
    logging.basicConfig(level=loglevel)
    logger.debug("loglevel set to '%s'", logging._levelToName[loglevel])


def run_with_arguments():
    """
    Main execution function.

    This is the function executed when yadg is launched via its executable.
    The function has the following subcommands:

    - ``extract``: extracts (meta)data from a given file.
    - ``process``: processes a given dataschema, extracting data into a NetCDF file.
    - ``update``: updates a given dataschema to the current version.
    - ``preset``: creates a dataschema from a preset file and a target folder.

    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--version",
        action="version",
        version=f'%(prog)s version {metadata.version("yadg")}',
    )

    verbose = argparse.ArgumentParser(add_help=False)

    for p in [parser, verbose]:
        p.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help="Increase verbosity by one level.",
        )
        p.add_argument(
            "--quiet",
            "-q",
            action="count",
            default=0,
            help="Decrease verbosity by one level.",
        )

    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    process = subparsers.add_parser("process")
    process.add_argument(
        "infile",
        help="File containing the dataschema to be processed by yadg.",
        default=None,
    )
    process.add_argument(
        "outfile",
        nargs="?",
        help="Output NetCDF file to save the created datatree to.",
        default="datagram.nc",
    )
    process.add_argument(
        "--ignore-merge-errors",
        dest="ignore_merge_errors",
        action="store_true",
        help="Ignore metadata merge errors while processing multiple files in a step.",
        default=False,
    )
    process.set_defaults(func=subcommands.process)

    update = subparsers.add_parser("update")
    update.add_argument(
        "infile",
        help="The file containing the dataschema in the old format.",
    )
    update.add_argument(
        "outfile",
        nargs="?",
        help="Output file to save the updated dataschema to.",
        default=None,
    )
    update.set_defaults(func=subcommands.update)

    preset = subparsers.add_parser("preset")
    preset.add_argument(
        "-p",
        "--process",
        action="store_true",
        help="Immediately process the dataschema created from the preset.",
        default=False,
    )
    preset.add_argument(
        "-a",
        "--archive",
        action="store_true",
        help="Archive the whole preset folder after processing.",
        default=False,
    )
    preset.add_argument(
        "--packwith",
        nargs="?",
        choices=["zip", "tar", "bztar", "xztar", "gztar"],
        help="Select compression algorithm for --archive.",
        default="zip",
    )
    preset.add_argument(
        "preset",
        help="Specify a preset, i.e. a dataschema template.",
    )
    preset.add_argument(
        "folder",
        help="Specify the folder on which to apply the preset.",
    )
    preset.add_argument(
        "outfile",
        nargs="?",
        help=(
            "Output file for the created dataschema file. Default is 'schema.json'. "
            "If '--process' is specified, the created datatree will be saved instead. "
            "Default in that case is 'datagram.nc'."
        ),
        default=None,
    )
    preset.add_argument(
        "--ignore-merge-errors",
        dest="ignore_merge_errors",
        action="store_true",
        help="Ignore metadata merge errors while processing multiple files in a step.",
        default=False,
    )
    preset.set_defaults(func=subcommands.preset)

    extract = subparsers.add_parser("extract")
    extract.add_argument(
        "filetype",
        help="Specify the filetype of the infile, selecting an appropriate extractor.",
    )
    extract.add_argument(
        "infile",
        help="Specify the input file which should be extracted.",
    )
    extract.add_argument(
        "outfile",
        nargs="?",
        help="Specify the output file name.",
        default=None,
    )
    extract.add_argument(
        "--meta-only",
        "-m",
        action="store_true",
        help="Extract and return file metadata only in a JSON format.",
        default=False,
    )
    extract.add_argument(
        "--locale",
        help="Set locale of the extracted file.",
        default=None,
        type=str,
    )
    extract.add_argument(
        "--encoding",
        help="Set encoding of the extracted file.",
        default=None,
        type=str,
    )
    extract.add_argument(
        "--timezone",
        help="Set timezone for the extracted file.",
        default=None,
        type=str,
    )
    extract.set_defaults(func=subcommands.extract)

    # parse subparser args
    args, extras = parser.parse_known_args()
    # parse extras for verbose tags
    args, extras = verbose.parse_known_args(extras, args)
    # assign last extra to outfile
    if len(extras) == 1:
        args.outfile = extras[0]

    set_loglevel(args.verbose - args.quiet)

    if "func" in args:
        args.func(**vars(args))



"""
Handles the reading and processing of volumetric flow meter data exported from the
MesaLabs DryCal software as a rtf file.

.. note::

    The date information is missing in the timestamps of the exported files and has to
    be supplied externally. The timestamp in the header of the rtf file corresponds to
    the timestamp of export / report generation, not measurement.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Drycal_rtf

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float               # Unix timestamp, without date
      data_vars:
        DryCal:         (uts)                 # Standardised flow rate
        DryCal Avg.:    (uts)                 # Running average of the flow rate
        Temp.:          (uts)                 # Measured flow temperature
        Pressure:       (uts)                 # Measured flow pressure

Metadata
````````
The following metadata is extracted:

    - ``product``: Model name of the MesaLabs device.
    - ``serial number``: Serial number of the MesaLabs device.

Uncertainties
`````````````
All uncertainties are derived from the string representation of the floats.

.. codeauthor::
    Peter Kraus

"""

from datatree import DataTree
from yadg.extractors.drycal import common


def extract(
    *,
    fn: str,
    encoding: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    vals = common.rtf(fn, encoding, timezone)
    return DataTree(common.check_timestamps(vals))



import pytest
import os
import pickle
from .utils import compare_datatrees, datagram_from_file


@pytest.mark.parametrize(
    "infile,",
    (
        "ts0.yml",
        "ts1.yml",
        "ts2.yml",
        "ts3.yml",
        "ts4.yml",
        "ts5.yml",
        "ts6.yml",
    ),
)
def test_externaldate_dataschema(infile, datadir):
    os.chdir(datadir)
    ret = datagram_from_file(infile)
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref)



import numpy as np
import xarray as xr
from xarray import Dataset
from typing import Any
import logging

logger = logging.getLogger(__name__)


def append_dicts(
    vals: dict[str, Any],
    devs: dict[str, Any],
    data: dict[str, list[Any]],
    meta: dict[str, list[Any]],
    fn: str = None,
    li: int = 0,
) -> None:
    if "_fn" in meta and fn is not None:
        meta["_fn"].append(str(fn))
    for k, v in vals.items():
        if k not in data:
            data[k] = [None if isinstance(v, str) else np.nan] * li
        data[k].append(v)
    for k, v in devs.items():
        if k not in meta:
            meta[k] = [np.nan] * li
        meta[k].append(v)

    for k in set(data) - set(vals):
        data[k].append(np.nan)
    for k in set(meta) - set(devs):
        if k != "_fn":
            meta[k].append(np.nan)


def dicts_to_dataset(
    data: dict[str, list[Any]],
    meta: dict[str, list[Any]],
    units: dict[str, str] = dict(),
    fulldate: bool = True,
) -> Dataset:
    darrs = {}
    for key, val in data.items():
        attrs = {}
        u = units.get(key, None)
        if u is not None:
            attrs["units"] = u
        if key == "uts":
            continue
        if "/" in key:
            logger.warning(f"Replacing '/' for '_' in column {key!r}.")
            k = key.replace("/", "_")
        else:
            k = key
        darrs[k] = xr.DataArray(data=val, dims=["uts"], attrs=attrs)
        if key in meta and darrs[k].dtype.kind in {"i", "u", "f", "c", "m", "M"}:
            err = f"{k}_std_err"
            darrs[k].attrs["ancillary_variables"] = err
            attrs["standard_name"] = f"{k} standard_error"
            darrs[err] = xr.DataArray(data=meta[key], dims=["uts"], attrs=attrs)
    if "uts" in data:
        coords = dict(uts=data.pop("uts"))
    else:
        coords = dict()
    if fulldate:
        attrs = dict()
    else:
        attrs = dict(fulldate=False)
    return xr.Dataset(data_vars=darrs, coords=coords, attrs=attrs)


def merge_dicttrees(vals: dict, fvals: dict, mode: str) -> dict:
    """
    A helper function that merges two ``DataTree.to_dict()`` objects by concatenating
    the new values in ``fvals`` to the existing ones in ``vals``.

    """
    if vals is None:
        return fvals
    for k in fvals.keys():
        try:
            vals[k] = xr.concat([vals[k], fvals[k]], dim="uts", combine_attrs=mode)
        except xr.MergeError:
            raise RuntimeError(
                "Merging metadata from multiple files has failed, as some of the "
                "values differ between files. This might be caused by trying to "
                "parse data obtained using different techniques/protocols in a "
                "single step. If you are certain this is what you want, try using "
                "yadg with the '--ignore-merge-errors' option."
                f"\n{vals[k].attrs=}"
                f"\n{fvals[k].attrs=}"
            )
    return vals


def merge_meta(old: dict, new: dict):
    for k, v in new.items():
        if k not in old:
            old[k] = v
        elif v == old[k]:
            pass
        elif isinstance(old[k], tuple):
            old[k] = (*old[k], v)
        else:
            old[k] = (old[k], v)



"""
Extractor of Agilent OpenLab binary signal trace files (``.ch`` and ``.it``).
Currently supports version "179" of the files. Version information is defined in
the ``magic_values`` (parameters & metadata) and `data_dtypes` (data) dictionaries.

Adapted from `ImportAgilent.m <https://bit.ly/3HSelIR>`_ and
`aston <https://github.com/bovee/Aston>`_.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Agilent_ch

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      {{ detector_name }}:
        coords:
          uts:            !!float               # Unix timestamp
          elution_time:   !!float               # Elution time
        data_vars:
          signal:         (uts, elution_time)   # Signal data

Metadata
````````
The following metadata is extracted:

    - ``sampleid``: Sample name.
    - ``username``: User name used to generate the file.
    - ``method``: Name of the chromatographic method.
    - ``version``: Version of the CH file (only "179" is currently supported.)


Notes on file structure
```````````````````````
The following magic values are used:
.. code ::

    0x0000 "version magic"
    0x0108 "data offset"
    0x011a "x-axis minimum (ms)"
    0x011e "x-axis maximum (ms)"
    0x035a "sample ID"
    0x0559 "description"
    0x0758 "username"
    0x0957 "timestamp"
    0x09e5 "instrument name"
    0x09bc "inlet"
    0x0a0e "method"
    0x104c "y-axis unit"
    0x1075 "detector name"
    0x1274 "y-axis intercept"
    0x127c "y-axis slope"

Data is stored in a consecutive set of ``<f8``, starting at the offset (calculated
as ``offset = ("data offset" - 1) * 512``) until the end of the file.

Uncertainties
`````````````
Uncertainty in ``signal`` is the y-axis slope.

Uncertainty in ``elution_time`` is the x-axis step size.

.. codeauthor::
    Peter Kraus

"""

import numpy as np
from yadg import dgutils
import xarray as xr
from datatree import DataTree

magic_values = {}
magic_values["179"] = {
    0x035A: ("sampleid", "utf-16"),
    0x0559: ("description", "utf-16"),
    0x0A0E: ("method", "utf-16"),
    0x0758: ("username", "utf-16"),
    0x0957: ("timestamp", "utf-16"),
    0x09E5: ("instrument", "utf-16"),
    0x09BC: ("inlet", "utf-16"),
    0x104C: ("yunit", "utf-16"),
    0x1075: ("tracetitle", "utf-16"),
    0x0108: ("offset", ">i4"),  # (x-1) * 512
    0x011A: ("xmin", ">f4"),  # / 60000
    0x011E: ("xmax", ">f4"),  # / 60000
    0x1274: ("intercept", ">f8"),
    0x127C: ("slope", ">f8"),
}

data_dtypes = {}
data_dtypes["179"] = (8, "<f8")


def extract(
    *,
    fn: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "rb") as inf:
        ch = inf.read()

    magic = dgutils.read_value(ch, 0, "utf-8")
    pars = {}
    orig_meta = {}
    if magic in magic_values.keys():
        for offset, (tag, dtype) in magic_values[magic].items():
            v = dgutils.read_value(ch, offset, dtype)
            orig_meta[tag] = v
    orig_meta["version"] = magic
    pars["end"] = len(ch)
    pars["start"] = (orig_meta["offset"] - 1) * 512
    dsize, ddtype = data_dtypes[magic]
    nbytes = pars["end"] - pars["start"]
    assert nbytes % dsize == 0
    npoints = nbytes // dsize

    xsn = np.linspace(orig_meta["xmin"] / 1000, orig_meta["xmax"] / 1000, num=npoints)
    xss = np.ones(npoints) * xsn[0]
    ysn = (
        np.frombuffer(
            ch,
            offset=pars["start"],
            dtype=ddtype,
            count=npoints,
        )
        * orig_meta["slope"]
    )
    yss = np.ones(npoints) * orig_meta["slope"]

    detector, title = orig_meta["tracetitle"].split(",")

    uts = dgutils.str_to_uts(
        timestamp=orig_meta["timestamp"], format="%d-%b-%y, %H:%M:%S", timezone=timezone
    )

    ds = xr.Dataset(
        data_vars={
            "signal": (
                ["uts", "elution_time"],
                [ysn],
                {"units": orig_meta["yunit"], "ancillary_variables": "signal_std_err"},
            ),
            "signal_std_err": (
                ["uts", "elution_time"],
                [yss],
                {"units": orig_meta["yunit"], "standard_name": "signal standard_error"},
            ),
            "elution_time_std_err": (
                ["elution_time"],
                xss,
                {"units": "s", "standard_name": "elution_time standard_error"},
            ),
        },
        coords={
            "elution_time": (
                ["elution_time"],
                xsn,
                {"units": "s", "ancillary_variables": "elution_time_std_err"},
            ),
            "uts": (["uts"], [uts]),
        },
        attrs=dict(original_metadata={"title": title}),
    )
    dt = DataTree.from_dict({detector: ds})
    dt.attrs = {"original_metadata": orig_meta}
    return dt



"""
This module parses the files generated by the ``dummy`` and ``biologic`` devices
within ``tomato-0.2``. As the ``dummy`` device has been mainly used for testing,
the below discusses the output of a ``biologic`` device.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Tomato_json

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float     # Unix timestamp
      data_vars:
        Ewe:            (uts)       # Potential of the working electrode
        Ece:            (uts)       # Potential of the counter electrode, if present
        I:              (uts)       # Instantaneous current
        technique:      (uts)       # Technique name
        loop number:    (uts)       # Loop number  (over techniques)
        cycle number:   (uts)       # Cycle number (within technique)
        index:          (uts)       # Technique index

Metadata
````````
No metadata is extracted.

Notes on file structure
```````````````````````
The files generated by the ``dummy`` driver do not contain the ``technique``, and all
values present in the json files are simply copied over assuming an uncertainty of 0.0.

For the ``biologic`` driver, each tomato data file contains the following four sections:

- ``technique`` section, describing the current technique,
- ``previous`` section, containing status information of the previous file,
- ``current`` section, containing status information of the current file,
- ``data`` section, containing the timesteps.

The reason why both ``previous`` and ``current`` are requires is that the device
status is recorded at the time of data polling, which means the values in ``current``
might be invalid (after the run has finished) or not in sync with the ``data`` (if
a technique change happened). However, ``previous`` may not be present in the first
data file of an experiment.

Uncertainties
`````````````
To determine the measurement errors, the values from BioLogic manual are used: for
measured voltages (:math:`E_{\\text{we}}` and :math:`E_{\\text{ce}}`) this corresponds
to a constant uncertainty of 0.004% of the applied E-range with a maximum of 75 uV,
while for currents (:math:`I`) this is a constant uncertainty of 0.0015% of the applied
I-range with a maximum of 0.76 uA.

.. codeauthor::
    Peter Kraus

"""

import json
import logging
import xarray as xr
from datatree import DataTree

from yadg import dgutils

logger = logging.getLogger(__name__)

I_ranges = {
    "1 A": 1e0,
    "100 mA": 1e-1,
    "10 mA": 1e-2,
    "1 mA": 1e-3,
    "100 uA": 1e-4,
    "10 uA": 1e-5,
    "1 uA": 1e-6,
    "100 pA": 1e-7,
}


def biologic_tomato_json(fn: str, jsdata: dict) -> DataTree:
    technique = jsdata["technique"]
    previous = jsdata.get("previous", None)
    current = jsdata["current"]

    if "uts" in technique:
        uts = technique["uts"]
        fulldate = True
    else:
        uts = 0
        fulldate = False

    uts += technique["start_time"]

    if previous is None:
        meta = current
    elif current["status"] == "STOP":
        meta = previous
    elif previous["elapsed_time"] > technique["start_time"]:
        meta = previous
    else:
        meta = current

    I_range = I_ranges[meta["I_range"]]
    E_range = meta["E_range"]["max"] - meta["E_range"]["min"]

    data_vars = {
        "loop number": [],
        "technique": [],
        "index": [],
        "time": [],
        "Ewe": [],
        "Ewe_std_err": [],
        "Ece": [],
        "Ece_std_err": [],
        "I": [],
        "I_std_err": [],
        "cycle": [],
    }

    for point in jsdata["data"]:
        for k, v in point.items():
            if k == "time":
                data_vars[k].append(uts + v)
            elif k in {"Ewe", "Ece"}:
                data_vars[k].append(v)
                data_vars[f"{k}_std_err"].append(max(E_range * 0.0015 / 100, 75e-6))
            elif k in {"I"}:
                data_vars[k].append(v)
                data_vars[f"{k}_std_err"].append(max(I_range * 0.004 / 100, 760e-12))
            elif k in {"cycle"}:
                data_vars[k].append(v)
            else:
                logger.cricital(f"parameter {k}: {v} not understood.")
        data_vars["loop number"].append(technique["loop_number"])
        data_vars["technique"].append(technique["name"])
        data_vars["index"].append(technique["index"])

    data_vars["cycle number"] = data_vars.pop("cycle")
    uts = data_vars.pop("time")

    data_vars = {k: v for k, v in data_vars.items() if len(v) > 0}

    for k in data_vars:
        if k in {"Ewe", "Ece", "I"}:
            data_vars[k] = (
                ["uts"],
                data_vars[k],
                {
                    "units": "A" if k == "I" else "V",
                    "ancillary_variables": f"{k}_std_err",
                },
            )
        elif k.endswith("_std_err"):
            data_vars[k] = (
                ["uts"],
                data_vars[k],
                {
                    "units": "A" if k == "I" else "V",
                    "standard_name": f"{k.replace('_std_err', '')} standard_error",
                },
            )
        else:
            data_vars[k] = (["uts"], data_vars[k])

    ds = xr.Dataset(data_vars, coords=dict(uts=uts))
    if not fulldate:
        ds.attrs["fulldate"] = False
    return DataTree(ds)


def dummy_tomato_json(fn: str, jsdata: dict) -> DataTree:
    data_vals = {}
    meta_vals = {}
    for vi, vals in enumerate(jsdata["data"]):
        vals["uts"] = vals.pop("time")
        devs = {}
        for k, v in vals.items():
            if k not in {"time", "address", "channel"}:
                devs[k] = 0.0
        dgutils.append_dicts(vals, devs, data_vals, meta_vals, fn, vi)
    return DataTree(dgutils.dicts_to_dataset(data_vals, meta_vals, fulldate=False))


def extract(
    *,
    fn: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r") as inf:
        jsdata = json.load(inf)

    if "technique" in jsdata:
        return biologic_tomato_json(fn, jsdata)
    else:
        return dummy_tomato_json(fn, jsdata)



"""
Handles processing of the xlsx version of the structured format produced by Agilent's
Online LC device at Empa. It contains three sections:

  - metadata section,
  - table containing sampling information,
  - table containing analysed chromatography data.

Usage
`````
Available since ``yadg-4.2``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.EmpaLC_xlsx

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float               # Unix timestamp
        species:        !!str                 # Species name
      data_vars:
        height:         (uts, species)        # Peak height
        area:           (uts, species)        # Integrated peak area
        concentration:  (uts, species)        # Peak area with calibration applied
        retention time: (uts, species)        # Position of peak maximum

Metadata
````````
The following metadata is extracted:

  - ``sequence``: Sample / sequence name.
  - ``description``: A free-form description of the experiment.
  - ``username``: User name used to generate the file.
  - ``datafile``: Original path of the result file.
  - ``version``: Version of the export function used to generate the file.

.. codeauthor::
    Peter Kraus

"""

import logging
import datetime
import openpyxl
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
import xarray as xr
from datatree import DataTree
import numpy as np

logger = logging.getLogger(__name__)


def extract(
    *,
    fn: str,
    **kwargs: dict,
) -> DataTree:
    try:
        wb = openpyxl.load_workbook(
            filename=fn,
            read_only=True,
        )
    except TypeError:
        raise RuntimeError(
            f"Could not read the file '{fn}' using openpyxl. Try to open and save the "
            f"file in Excel."
        )

    ws = wb["Page 1"]
    metadata = {}
    for row in ws.rows:
        val = row[1].value if len(row) > 1 and row[1].value is not None else ""
        if row[0].value.startswith("Sequence name"):
            metadata["sequence"] = val
        elif row[0].value.startswith("Description"):
            metadata["description"] = val
        elif row[0].value.startswith("Acquired by"):
            metadata["username"] = val
        elif row[0].value.startswith("Data path"):
            metadata["datafile"] = val
        elif row[0].value.startswith("Report version"):
            metadata["version"] = int(val)

    if metadata.get("version", None) is None:
        raise RuntimeError(f"Report version in file '{fn}' was not specified.")

    ws = wb["Page 2"]
    samples = {}
    for row in ws.rows:
        if "Line#" in row[0].value:
            headers = [i.value.replace("\n", "").replace(" ", "") for i in row]
        else:
            data = [str(i.value) if i.value is not None else None for i in row]
            sample = {
                "location": data[headers.index("Location")],
                "injection date": data[headers.index("InjectionDate")],
                "acquisition": {
                    "method": data[headers.index("AcqMethodName")],
                    "version": data[headers.index("AcqMethodVersion")],
                },
                "integration": {
                    "method": data[headers.index("InjectionDAMethodName")],
                    "version": data[headers.index("InjectionDAMethodVersion")],
                },
                "offset": data[headers.index("Timeoffset")],
            }
            if sample["offset"] is not None:
                sn = data[headers.index("SampleName")]
                sn = sn.replace(" ", "").replace("\n", "")
                samples[sn] = sample

    svals = samples.values()
    if len(svals) == 0:
        raise RuntimeError(
            f"No complete sample data found in file '{fn}'. "
            "Have you added time offsets?"
        )
    r = next(iter(svals))
    # check that acquisition and integration methods are consistent throughout file:
    if any([s["acquisition"]["method"] != r["acquisition"]["method"] for s in svals]):
        logger.warning("Acquisition method is inconsistent in file '%s'.", fn)
    if any([s["integration"]["method"] != r["integration"]["method"] for s in svals]):
        logger.warning("Integration method is inconsistent in file '%s'.", fn)

    metadata["method"] = r["acquisition"]["method"].replace("\n", "").replace(" ", "")

    species = set()
    ws = wb["Page 3"]
    for row in ws.rows:
        if "Line#" in str(row[0].value):
            headers = [i.value.replace("\n", "").replace(" ", "") for i in row]
        else:
            data = [str(i.value) if i.value is not None else None for i in row]
            sn = data[headers.index("SampleName")].replace("\n", "").replace(" ", "")
            cn = data[headers.index("Compound")]
            species.add(cn)

            h = data[headers.index("PeakHeight")]
            if h is not None:
                if "height" not in samples[sn]:
                    samples[sn]["height"] = {}
                samples[sn]["height"][cn] = tuple_fromstr(h)

            A = data[headers.index("Area")]
            if A is not None:
                if "area" not in samples[sn]:
                    samples[sn]["area"] = {}
                samples[sn]["area"][cn] = tuple_fromstr(A)

            if metadata["version"] == 2:
                c = data[headers.index("Concentration")]
            else:
                logger.warning(
                    "Report version '%d' in file '%s' not understood.",
                    metadata["version"],
                    fn,
                )
                c = data[headers.index("Concentration")]
            if c is not None:
                if "concentration" not in samples[sn]:
                    samples[sn]["concentration"] = {}
                samples[sn]["concentration"][cn] = tuple_fromstr(c)

            rt = data[headers.index("RT[min]")]
            if rt is not None:
                if "retention time" not in samples[sn]:
                    samples[sn]["retention time"] = {}
                samples[sn]["retention time"][cn] = tuple_fromstr(rt)

    units = {
        "height": None,
        "area": None,
        "concentration": "mmol/l",
        "retention time": "min",
    }
    species = sorted(species)
    data = []
    for k, v in samples.items():
        # Remove unnecessary parameters
        del v["acquisition"]
        del v["integration"]
        v["sampleid"] = k
        # Process offset to uts
        offset = v.pop("offset")
        t = None
        for fmt in {"%H:%M:%S"}:
            try:
                t = datetime.datetime.strptime(offset, fmt)
            except ValueError:
                continue
        if t is None:
            try:
                td = datetime.timedelta(minutes=float(offset))
            except ValueError:
                raise RuntimeError(
                    f"It was not possible to parse offset '{offset}' present in file "
                    f"'{fn}' using known formats."
                )
        else:
            td = datetime.timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)
        point = {"uts": td.total_seconds()}
        vals = {}
        devs = {}
        for kk in {"height", "area", "concentration", "retention time"}:
            val = v.get(kk, {})
            vals[kk], devs[kk] = zip(*[val.get(cn, (np.nan, np.nan)) for cn in species])
        point["vals"] = vals
        point["devs"] = devs
        data.append(point)

    data_vars = {}
    for kk in {"height", "area", "concentration", "retention time"}:
        data_vars[kk] = (
            ["uts", "species"],
            [i["vals"][kk] for i in data],
            {"anciliary_variables": f"{kk}_std_err"},
        )
        data_vars[f"{kk}_std_err"] = (
            ["uts", "species"],
            [i["devs"][kk] for i in data],
            {"standard_name": f"{kk} standard_error"},
        )
        if units[kk] is not None:
            data_vars[kk][2]["units"] = units[kk]
            data_vars[f"{kk}_std_err"][2]["units"] = units[kk]

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={
            "species": (["species"], species),
            "uts": (["uts"], [i["uts"] for i in data]),
        },
        attrs=dict(original_metadata=metadata),
    )

    return DataTree(ds)



import pytest
import os
import pickle
import xarray as xr
from yadg.extractors.eclab.mpt import extract
from .utils import compare_datatrees


@pytest.mark.parametrize(
    "infile, locale",
    [
        ("ca.mpt", "en_GB"),
        ("ca.issue_134.mpt", "en_GB"),
        ("ca.issue_149.mpt", "de_DE"),
        ("cp.mpt", "en_GB"),
        ("cp.issue_61.mpt", "en_GB"),
        ("cp.issue_149.mpt", "de_DE"),
        ("cv.mpt", "en_GB"),
        ("cv.issue_149.mpt", "de_DE"),
        ("cva.issue_135.mpt", "en_GB"),
        ("gcpl.mpt", "en_GB"),
        ("gcpl.issue_149.mpt", "de_DE"),
        ("geis.mpt", "en_GB"),
        ("geis.issue_149.mpt", "de_DE"),
        ("lsv.mpt", "en_GB"),
        ("mb.mpt", "en_GB"),
        ("mb.issue_95.mpt", "en_GB"),
        ("mb.issue_149.mpt", "de_DE"),
        ("ocv.mpt", "en_GB"),
        ("ocv.issue_149.mpt", "de_DE"),
        ("peis.mpt", "en_GB"),
        ("peis.issue_149.mpt", "de_DE"),
        ("wait.mpt", "en_GB"),
        ("zir.mpt", "en_GB"),
        ("vsp_ocv_with.mpt", "en_GB"),
        ("vsp_ocv_wo.mpt", "en_GB"),
        ("vsp_peis_with.mpt", "en_GB"),
    ],
)
def test_eclab_mpt(infile, locale, datadir):
    os.chdir(datadir)
    ret = extract(
        fn=infile,
        timezone="Europe/Berlin",
        encoding="windows-1252",
        locale=locale,
    )
    print(f"{ret=}")
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)


@pytest.mark.parametrize(
    "afile, bfile",
    [
        ("mb.issue_95.mpt", "mb.issue_95.de.mpt"),
    ],
)
def test_eclab_mpt_locale(afile, bfile, datadir):
    os.chdir(datadir)
    kwargs = dict(timezone="Europe/Berlin", encoding="windows-1252")
    aret = extract(fn=afile, locale="en_US", **kwargs)
    bret = extract(fn=bfile, locale="de_DE", **kwargs)
    compare_datatrees(aret, bret)


@pytest.mark.parametrize(
    "afile, bfile",
    [
        ("mb.old.mpt", "mb.mpt"),
        ("gcpl.old.mpt", "gcpl.mpt"),
        ("ocv.old.mpt", "ocv.mpt"),
        ("wait.old.mpt", "wait.mpt"),
    ],
)
def test_eclab_mpt_old(afile, bfile, datadir):
    os.chdir(datadir)
    kwargs = dict(timezone="Europe/Berlin", encoding="windows-1252")
    aret = extract(fn=afile, locale="en_US", **kwargs)
    bret = extract(fn=bfile, locale="en_US", **kwargs)

    for key in aret.variables:
        try:
            xr.testing.assert_allclose(aret[key], bret[key])
        except AssertionError as e:
            if key in {"Efficiency", "Efficiency_std_err"}:
                pass
            else:
                e.args = (e.args[0] + f"Error happened on key: {key!r}\n",)
                raise e



from packaging.version import Version
import logging
import os
from typing import Union
from dgbowl_schemas.yadg import to_dataschema
from dgbowl_schemas.yadg.dataschema import DataSchema
from pydantic import BaseModel
from pydantic.v1 import BaseModel as BaseModel_v1
from yadg import dgutils

__latest_dataschema__ = "5.1"
logger = logging.getLogger(__name__)


def calib_3to4(oldcal: dict, caltype: str) -> dict:
    newcal = {}
    if caltype == "calfile":
        for k, v in oldcal["detectors"].items():
            pd = {
                "window": (v.get("window", 3) - 1) // 2,
                "polyorder": v.get("poly", 2),
                "prominence": v.get("prominence", 1.0),
                "threshold": v.get("threshold", 1.0),
            }
            sp = {}
            for kk, vv in v["species"].items():
                if kk == "units" and vv == "min":
                    continue
                spec = {
                    "l": vv["l"] * 60.0,
                    "r": vv["r"] * 60.0,
                    "calib": {"inverse": {"slope": vv.get("rf", 1.0)}},
                }
                sp[kk] = spec
            id = {"det_1": 0, "det_2": 1}[v["id"]]
            newcal[k] = {"id": id, "peakdetect": pd, "species": sp}
    elif caltype == "Tcalfile":
        newcal = {"T": {"T_f": {"calib": {"linear": oldcal}}, "unit": "degC"}}
    elif caltype == "MFCcalfile":
        newcal = {}
        for k, v in oldcal.items():
            items = v.get("content", {k: 1.0})
            for kk, vv in items.items():
                if kk not in newcal:
                    newcal[kk] = {"unit": "ml/min"}
                newcal[kk][k] = {
                    "calib": {
                        "linear": {
                            "slope": v.get("slope", 1.0),
                            "intercept": v.get("intercept", 0.0),
                        }
                    },
                    "fraction": vv,
                }
    return newcal


def schema_3to4(oldschema: list) -> dict:
    newschema = {
        "metadata": {
            "provenance": {
                "type": "yadg update",
                "metadata": {
                    "yadg": dgutils.get_yadg_metadata(),
                    "update_schema": {"updater": "schema_3to4"},
                },
            },
            "version": "4.1",
            "timezone": "localtime",
        },
        "steps": [],
    }
    for oldstep in oldschema:
        newstep = {}

        newstep["parser"] = oldstep["datagram"]
        if newstep["parser"] == "gctrace":
            newstep["parser"] = "chromtrace"

        if "paths" in oldstep["import"]:
            oldstep["import"]["files"] = oldstep["import"].pop("paths")
        newstep["input"] = oldstep["import"]

        if oldstep.get("export", None) is not None:
            newstep["tag"] = oldstep["export"]

        parameters = {}
        for k, v in oldstep["parameters"].items():
            if k in ["Tcalfile", "MFCcalfile", "calfile"]:
                logger.warning(
                    "Parsing of post-processing parameter '%s' has been removed in "
                    "yadg-5.0, please use dgpost-2.0 to reproduce this functionality.",
                    k,
                )
            elif k == "method" and v == "q0refl":
                logger.warning(
                    "Parsing of post-processing parameter '%s' has been removed in "
                    "yadg-5.0, please use dgpost-2.0 to reproduce this functionality.",
                    k,
                )
            else:
                parameters[k] = v
        if parameters != {}:
            newstep["parameters"] = parameters
        newschema["steps"].append(newstep)

    return newschema


def update_schema(object: Union[list, dict, BaseModel, BaseModel_v1]) -> DataSchema:
    """
    The ``yadg update`` worker function.

    The main purpose is to allow a simple update pathway from older versions of
    dataschema files to the current latest and greatest.

    Currently supports:

     - updating dataschema version 3.1 to 4.0 using routines in ``yadg``,
     - updating dataschema version 4.0 and above to the latest dataschema using the
       in-built ``.update()`` mechanism.

    Parameters
    ----------
    object
        The object to be updated

    Returns
    -------
    newobj: dict
        The updated and validated `"datagram"` or `"schema"`.

    """

    if isinstance(object, list):
        logger.info("Updating list-style DataSchema")
        newobj = to_dataschema(**schema_3to4(object))
    elif isinstance(object, dict):
        logger.info("Updating dict-style DataSchema")
        newobj = to_dataschema(**object)
    elif isinstance(object, (BaseModel, BaseModel_v1)):
        logger.info("Updating an existing DataSchema object")
        newobj = object
    else:
        raise ValueError(f"Supplied object is of incorrect type: {type(object)}")

    maxver = Version(__latest_dataschema__)
    while hasattr(newobj, "update"):
        temp = newobj.update()
        if hasattr(temp, "version"):
            if Version(temp.version) > maxver:
                break
        newobj = temp
    return newobj


def schema_from_preset(preset: DataSchema, folder: str) -> DataSchema:
    # preset.metadata["provenance"] = "yadg preset"
    for step in preset.steps:
        for fi, fn in enumerate(step.input.files):
            if os.path.isabs(fn):
                logger.warning(
                    "Item '%s' in '%s' is an absolute path and will not be patched.",
                    fn,
                    step,
                )
            else:
                step.input.files[fi] = os.path.abspath(os.path.join(folder, fn))
        if (
            step.externaldate is not None
            and hasattr(step.externaldate.using, "file")
            and step.externaldate.using.file is not None
        ):
            oldf = step.externaldate.using.file.path
            if os.path.isabs(oldf):
                logger.warning(
                    "Specified externaldate file '%s' is an absolute path "
                    "and will not be patched.",
                    oldf,
                )
            else:
                newf = os.path.abspath(os.path.join(folder, oldf))
                step.externaldate.using.file.path = newf
    return preset



"""
The `sac2dat.c code from Dr. Moritz Bubek <https://www.bubek.org/sac2dat.php>`_
was a really useful stepping stone for this Python file parser.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Quadstar_sac

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      {{ trace_index }}:
        coords:
          uts:              !!float               # Unix timestamp
          mass_to_charge:   !!float               # M/Z ratio
        data_vars:
          fsr:              (None)                # Full scale range
          y:                (uts, mass_to_charge) # Signal data


Metadata
````````
The "info_position" section in the below structure is stored as metadata for every
trace, without further processing.

Notes on file structure
```````````````````````
Pretty much the entire file format has been reverse engineered. There
are still one or two unknown fields.

.. code-block:: python

    0x00 "data_index"
    0x02 "software_id"
    0x06 "version_major"
    0x07 "version_minor"
    0x08 "second"
    0x09 "minute"
    0x0a "hour"
    0x0b "day"
    0x0c "month"
    0x0d "year"
    0x0f "author"
    0x64 "n_timesteps"
    0x68 "n_traces"
    0x6a "timestep_length"
    ...
    # Not sure what sits from 0x6e to 0xc2.
    ...
    0xc2 "uts_base_s"
    0xc6 "uts_base_ms"
    # Trace header. Read these 9 bytes for every trace (n_traces).
    0xc8 + (n * 0x09) "type"
    0xc9 + (n * 0x09) "info_position"
    0xcd + (n * 0x09) "data_position"
    ...
    # Trace info. Read these 137 bytes for every trace where type != 0x11.
    info_position + 0x00 "data_format"
    info_position + 0x02 "y_title"
    info_position + 0x0f "y_unit"
    info_position + 0x1d "x_title"
    info_position + 0x2a "x_unit"
    info_position + 0x38 "comment"
    info_position + 0x7a "first_mass"
    info_position + 0x7e "scan_width"
    info_position + 0x80 "values_per_mass"
    info_position + 0x81 "zoom_start"
    info_position + 0x85 "zoom_end"
    ...
    # UTS offset. Read these 6 bytes for every timestep (n_timesteps).
    0xc2 + (n * timestep_length) "uts_offset_s"
    0xc6 + (n * timestep_length) "uts_offset_ms"
    # Read everything remaining below for every timestep and every trace
    # where type != 0x11.
    data_position + (n * timestep_length) + 0x00 "n_datapoints"
    data_position + (n * timestep_length) + 0x04 "data_range"
    # Datapoints. Read these 4 bytes (scan_width * values_per_mass)
    # times.
    data_position + (n * timestep_length) + 0x06 "datapoints"
    ...

Uncertainties
`````````````
Uncertainties in ``mass_to_charge`` are set to one step in M/Z spacing.

Uncertainties in the signal ``y`` are either based on the analog-to-digital conversion
(i.e. using the full scale range), or from the upper limit of contribution of
neighboring M/Z points (50 ppm).

.. codeauthor::
    Nicolas Vetsch

"""

import numpy as np
from datatree import DataTree
import xarray as xr
import yadg.dgutils as dgutils

# The general header at the top of .sac files.
general_header_dtype = np.dtype(
    [
        ("data_index", "<i2"),
        ("software_id", "<i4"),
        ("version_major", "|u1"),
        ("version_minor", "|u1"),
        ("S", "|u1"),
        ("M", "|u1"),
        ("H", "|u1"),
        ("d", "|u1"),
        ("m", "|u1"),
        ("y", "|u1"),
        ("username", "|S86"),
        ("n_timesteps", "<i4"),
        ("n_traces", "<i2"),
        ("timestep_length", "<i4"),
    ]
)


trace_header_dtype = np.dtype(
    [
        ("type", "|u1"),
        ("info_position", "<i4"),
        ("data_position", "<i4"),
    ]
)


trace_info_dtype = np.dtype(
    [
        ("data_format", "<u2"),
        ("y_title", "|S13"),
        ("y_unit", "|S13"),
        ("unknown_a", "|u1"),
        ("x_title", "|S13"),
        ("x_unit", "|S13"),
        ("comment", "|S59"),
        ("unknown_b", "|u4"),
        ("unknown_c", "|u4"),
        ("first_mass", "<f4"),
        ("scan_width", "<u2"),
        ("values_per_mass", "|u1"),
        ("zoom_start", "<f4"),
        ("zoom_end", "<f4"),
    ]
)


def _find_first_data_position(scan_headers: list[dict]) -> int:
    """Finds the data position of the first scan containing any data."""
    for header in scan_headers:
        if header["type"] != 0x11:
            continue
        return header["data_position"]


def extract(
    *,
    fn: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "rb") as sac_file:
        sac = sac_file.read()
    meta = dgutils.read_value(sac, 0x0000, general_header_dtype)
    uts_base_s = dgutils.read_value(sac, 0x00C2, "<u4")
    # The ms part of the timestamps is actually saved as tenths of ms so
    # multiplying this by 0.1 here.
    uts_base_ms = dgutils.read_value(sac, 0x00C6, "<u2") * 1e-1
    uts_base = uts_base_s + uts_base_ms * 1e-3
    trace_headers = np.frombuffer(
        sac,
        offset=0x00C8,
        dtype=trace_header_dtype,
        count=meta["n_traces"],
    )
    # Find the data position of the first data-containing timestep.
    data_pos_0 = _find_first_data_position(trace_headers)
    traces = {}
    for n in range(meta["n_timesteps"]):
        ts_offset = n * meta["timestep_length"]
        uts_offset_s = dgutils.read_value(sac, data_pos_0 - 0x0006 + ts_offset, "<u4")
        uts_offset_ms = (
            dgutils.read_value(sac, data_pos_0 - 0x0002 + ts_offset, "<u2") * 1e-1
        )
        uts_timestamp = uts_base + (uts_offset_s + uts_offset_ms * 1e-3)
        for ti, header in enumerate(trace_headers):
            if header["type"] != 0x11:
                continue
            info = dgutils.read_value(sac, header["info_position"], trace_info_dtype)
            # Construct the mass data.
            ndm = info["values_per_mass"]
            mvals, dm = np.linspace(
                info["first_mass"],
                info["first_mass"] + info["scan_width"],
                info["scan_width"] * ndm,
                endpoint=False,
                retstep=True,
            )
            mdevs = np.ones(len(mvals)) * dm
            # Read and construct the y data.
            ts_data_pos = header["data_position"] + ts_offset
            # Determine the detector's full scale range.
            fsr = 10 ** dgutils.read_value(sac, ts_data_pos + 0x0004, "<i2")
            # The n_datapoints value at timestep_data_position is
            # sometimes wrong. Calculating this here works, however.
            n_datapoints = info["scan_width"] * info["values_per_mass"]
            yvals = np.frombuffer(
                sac, offset=ts_data_pos + 0x0006, dtype="<f4", count=n_datapoints
            ).copy()
            # Once a y_value leaves the FSR it jumps to the maximum
            # of a float32. These values should be NaNs instead.
            yvals[yvals > fsr] = np.nan
            # TODO: Determine the correct accuracy from fsr. The 32bit
            # ADC is a guess that seems to put the error in the correct
            # order of magnitude.
            sigma_adc = np.ones(len(yvals)) * fsr / 2**32
            # Determine error based on contributions of neighboring masses.
            # The upper limit on contribution from peak at next integer mass is 50ppm.
            prev_neighbor = np.roll(yvals, ndm)
            prev_neighbor[:ndm] = np.nan
            next_neighbor = np.roll(yvals, -ndm)
            next_neighbor[-ndm:] = np.nan
            sigma_neighbor = np.fmax(prev_neighbor, next_neighbor) * 50e-6
            # Pick the maximum error here
            ydevs = np.fmax(sigma_adc, sigma_neighbor)
            ds = xr.Dataset(
                data_vars={
                    "fsr": fsr,
                    "mass_to_charge_std_err": (
                        ["mass_to_charge"],
                        mdevs,
                        {
                            "units": info["x_unit"],
                            "standard_name": "mass_to_charge standard_error",
                        },
                    ),
                    "y": (
                        ["uts", "mass_to_charge"],
                        [yvals],
                        {"units": info["y_unit"], "ancilliary_variables": "y_std_err"},
                    ),
                    "y_std_err": (
                        ["uts", "mass_to_charge"],
                        [ydevs],
                        {"units": info["y_unit"], "standard_name": "y standard_error"},
                    ),
                },
                coords={
                    "mass_to_charge": (
                        ["mass_to_charge"],
                        mvals,
                        {
                            "units": info["x_unit"],
                            "ancillary_variables": "mass_to_charge_std_err",
                        },
                    ),
                    "uts": (["uts"], [uts_timestamp]),
                },
                attrs=dict(original_metadata=info),
            )
            if f"{ti}" not in traces:
                traces[f"{ti}"] = ds
            else:
                try:
                    traces[f"{ti}"] = xr.concat(
                        [traces[f"{ti}"], ds], dim="uts", combine_attrs="identical"
                    )
                except xr.MergeError:
                    raise RuntimeError(
                        "Merging metadata from the individual traces has failed. "
                        "This is a bug. Please open an issue on GitHub."
                    )

    ret = DataTree.from_dict(traces)
    ret.attrs = dict(original_metadata=meta)
    return ret



"""
``pint`` compatibility functions in **yadg**.

This package defines ``ureg``, a :class:`pint.UnitRegistry` used for validation of
`datagrams` in **yadg**. The default SI :class:`pint.UnitRegistry` is extended
by definitions of fractional quantities (%, ppm, etc.), standard volumetric
quantities (smL/min, sccm), and other dimensionless "units" present in several
file types.
"""

import logging
from typing import Union
import pint

logger = logging.getLogger(__name__)

ureg = pint.UnitRegistry(
    preprocessors=[
        lambda s: s.replace("%%", " permille "),
        lambda s: s.replace("%", " percent "),
    ]
)

definitions = """
parts_per_hundred = 1e-2 count = percent
parts_per_thousand = 1e-3 count = ‰ = promile = permille
parts_per_million = 1e-6 count = ppm
parts_per_billion = 1e-9 count = ppb

standard_milliliter = milliliter = smL
standard_liter = liter = sL
standard_cubic_centimeter_minute = cm^3/min = sccm

refractive_index_units = [] = RIU
"""

for line in definitions.split("\n"):
    line = line.strip()
    if len(line) > 0 and not line.startswith("#"):
        ureg.define(line)


def _sanitize_helper(unit: str) -> str:
    unit = unit.replace("Bar", "bar")
    if unit in ["deg C", "Deg C"]:
        return "degC"
    else:
        return unit


def sanitize_units(
    units: Union[str, dict[str, str], list[str]],
) -> Union[str, dict[str, str], list[str]]:
    """
    Unit sanitizer.

    This sanitizer should be used where user-supplied units are likely to occur,
    such as in the parsers :mod:`yadg.extractors.basic.csv`. Currently, only two
    replacements are done:

      - "Bar" is replaced with "bar"
      - "Deg C" is replace with "degC

    Use with caution.

    Parameters
    ----------
    units
        Object containing string units.

    """
    if isinstance(units, list):
        for i in range(len(units)):
            units[i] = _sanitize_helper(units[i])
    elif isinstance(units, dict):
        for k, v in units.items():
            units[k] = _sanitize_helper(v)
    elif isinstance(units, str):
        units = _sanitize_helper(units)
    else:
        logger.error("Supplied type of 'units' not understood: '%s'", str(type(units)))
    return units



import pytest
import os
import pickle
from yadg.extractors.eclab.mpr import extract
from .utils import compare_datatrees


@pytest.mark.parametrize(
    "infile",
    [
        "ca.mpr",
        "ca.issue_134.mpr",
        "ca.issue_149.mpr",
        "cp.mpr",
        "cp.issue_61.mpr",
        "cp.issue_149.mpr",
        "cv.mpr",
        "cv.issue_149.mpr",
        "cva.issue_135.mpr",
        "gcpl.mpr",
        "gcpl.issue_149.mpr",
        "gcpl.issue_175.mpr",
        "geis.mpr",
        "geis.issue_149.mpr",
        "lsv.mpr",
        "mb.mpr",
        "mb.issue_95.mpr",
        "mb.issue_149.mpr",
        "ocv.mpr",
        "ocv.issue_149.mpr",
        "peis.mpr",
        "peis.issue_149.mpr",
        "wait.mpr",
        "zir.mpr",
        "vsp_ocv_with.mpr",
        "vsp_ocv_wo.mpr",
        "vsp_peis_with.mpr",
    ],
)
def test_eclab_mpr(infile, datadir):
    os.chdir(datadir)
    ret = extract(fn=infile, timezone="Europe/Berlin")
    print(f"{ret=}")
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)



import os
from .utils import datagram_from_file


def test_timestamp_parsing(datadir):
    os.chdir(datadir)
    utc = datagram_from_file("ssv_utc.json")
    cet = datagram_from_file("ssv_cet.json")
    diff = utc["0"].uts.to_numpy() - cet["0"].uts.to_numpy()
    assert diff[0] == 7200, "CEST should be 2 hours ahead of UTC"
    assert diff[1] == 7200, "CEST should be 2 hours ahead of UTC"
    assert diff[2] == 3600, "CET should be 1 hours ahead of UTC"


def test_uts_parsing(datadir):
    os.chdir(datadir)
    utc = datagram_from_file("csv_utc.json")
    cet = datagram_from_file("csv_cet.json")
    diff = utc["0"].uts.to_numpy() - cet["0"].uts.to_numpy()
    assert (diff == 0).all(), "'uts' column parsing should ignore timezones"


def test_isotimestamp_parsing(datadir):
    os.chdir(datadir)
    utc = datagram_from_file("csv_utc.json")
    cet = datagram_from_file("csv_cet.json")
    diff = utc["0"].uts.to_numpy() - cet["0"].uts.to_numpy()
    assert (diff == 0).all(), "Z-suffix should override timezone provided in schema"


def test_isotimestamp_parsing_utc(datadir):
    os.chdir(datadir)
    dg = datagram_from_file("iso_utc.json")

    assert dg["0"].uts[0] == 1622557825.0, "Z-suffix parsed to 'uts' correctly"
    assert (
        dg["0"].uts[1] == 1622557825.0 + 7200 + 1
    ), "±HH:MM-suffix parsed to 'uts' correctly"
    assert (
        dg["0"].uts[2] == 1622557825.0 + 2
    ), "no suffix with timezone = UTC parsed to 'uts' correctly"


def test_isotimestamp_parsing_cet(datadir):
    os.chdir(datadir)
    dg = datagram_from_file("iso_cet.json")
    assert dg["0"].uts[0] == 1622557825.0, "Z-suffix parsed to 'uts' correctly"
    assert (
        dg["0"].uts[1] == 1622557825.0 + 7200 + 1
    ), "±HH:MM-suffix parsed to 'uts' correctly"
    assert (
        dg["0"].uts[2] == 1622557825.0 - 7200 + 2
    ), "no suffix with timezone = CEST parsed to 'uts' correctly"



import pytest
import os
from yadg.extractors import extract
import datatree
from .utils import compare_datatrees


@pytest.mark.parametrize(
    "filetype, infile",
    [
        ("biologic-mpr", "cp.mpr"),
        ("biologic-mpt", "cp.mpt"),
        ("agilent-ch", "hplc.CH"),
        ("agilent-dx", "hplc.dx"),
        ("phi-spe", "xps.spe"),
        ("panalytical-xrdml", "xrd.xrdml"),
        ("touchstone.snp", "picovna.s1p"),
        ("ezchrom.asc", "230324.dat.asc"),
        ("ezchrom.dat", "230324.dat"),
        ("picolog.tc08", "20220723-porosity-study-15p-Cu-200mA-longrun-07.picolog"),
    ],
)
def test_yadg_extractors_extract_with_metadata(filetype, infile, datadir):
    os.chdir(datadir)
    outfile = f"{infile}.nc"
    ret = extract(filetype=filetype, path=infile)
    # ret.to_netcdf(f"C:/Users/Kraus/Code/yadg/tests/test_extract/{outfile}", engine="h5netcdf")
    ref = datatree.open_datatree(outfile, engine="h5netcdf")
    compare_datatrees(ret, ref, thislevel=True, descend=True)



"""
Extractor of Agilent OpenLab DX archives. This is a wrapper parser which unzips the
provided DX file, and then uses the :mod:`yadg.extractors.agilent.ch` extractor
to parse every CH file present in the archive. The IT files in the archive are currently
ignored.

.. note::

   Currently the timesteps from multiple CH files (if present) are appended in the
   timesteps array without any further sorting.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Agilent_dx

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      {{ detector_name }}:
        coords:
          uts:            !!float               # Unix timestamp
          elution_time:   !!float               # Elution time
        data_vars:
          signal:         (uts, elution_time)   # Signal data

Metadata
````````
The following metadata is extracted:

    - ``sampleid``: Sample name.
    - ``username``: User name used to generate the file.
    - ``method``: Name of the chromatographic method.
    - ``version``: Version of the CH file (only "179" is currently supported.)

.. codeauthor::
    Peter Kraus

"""

import zipfile
import tempfile
import os
from datatree import DataTree

from yadg.extractors.agilent.ch import extract as extract_ch
from yadg import dgutils


def extract(
    *,
    fn: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    zf = zipfile.ZipFile(fn)
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        dt = None
        filenames = [ffn for ffn in os.listdir(tempdir) if ffn.endswith("CH")]
        for ffn in sorted(filenames):
            path = os.path.join(tempdir, ffn)
            fdt = extract_ch(fn=path, timezone=timezone, **kwargs)
            dt = dgutils.merge_dicttrees(dt, fdt.to_dict(), "identical")
    return DataTree.from_dict(dt)



import pytest
import os
import pickle
from yadg.extractors.basic.csv import extract
from dgbowl_schemas.yadg.dataschema_5_1.filetype import Basic_csv
from .utils import compare_datatrees


@pytest.mark.parametrize(
    "infile, params",
    [
        (
            "case_uts_units.csv",
            {"timestamp": {"uts": {"index": 0}}},
        ),
        (
            "case_timestamp.ssv",
            {"sep": ";", "units": {"flow": "ml/min", "T": "K", "p": "atm"}},
        ),
        (
            "case_custom_ts.tsv",
            {
                "sep": "\t",
                "timestamp": {
                    "timestamp": {"index": 1, "format": "%d.%m.%Y %I:%M:%S%p"}
                },
            },
        ),
        (
            "picolog_temperature.csv",
            {
                "sep": ",",
                "timestamp": {"uts": {"index": 0}},
                "strip": '"',
                "units": {
                    "Outside Last (C)": "degC",
                    "Outside Ave. (C)": "degC",
                    "Outside Min. (C)": "degC",
                    "Outside Max. (C)": "degC",
                    "Inside_GDE_TCK01 Last (C)": "degC",
                    "Inside_GDE_TCK01 Ave. (C)": "degC",
                    "Inside_GDE_TCK01 Min. (C)": "degC",
                    "Inside_GDE_TCK01 Max. (C)": "degC",
                    "Inside_Nafion_TCK02 Last (C)": "degC",
                    "Inside_Nafion_TCK02 Ave. (C)": "degC",
                    "Inside_Nafion_TCK02 Min. (C)": "degC",
                    "Inside_Nafion_TCK02 Max. (C)": "degC",
                },
            },
        ),
        (
            "picolog_temperature_sparse.csv",
            {
                "sep": ",",
                "timestamp": {"uts": {"index": 0}},
                "strip": '"',
                "units": {
                    "Outside Last (C)": "degC",
                    "Outside Ave. (C)": "degC",
                    "Outside Min. (C)": "degC",
                    "Outside Max. (C)": "degC",
                    "Inside_GDE_TCK01 Last (C)": "degC",
                    "Inside_GDE_TCK01 Ave. (C)": "degC",
                    "Inside_GDE_TCK01 Min. (C)": "degC",
                    "Inside_GDE_TCK01 Max. (C)": "degC",
                    "Inside_Nafion_TCK02 Last (C)": "degC",
                    "Inside_Nafion_TCK02 Ave. (C)": "degC",
                    "Inside_Nafion_TCK02 Min. (C)": "degC",
                    "Inside_Nafion_TCK02 Max. (C)": "degC",
                },
            },
        ),
        (
            "flow_data.csv",
            {
                "sep": ",",
                "timestamp": {"time": {"index": 5}},
                "units": {
                    "DryCal smL/min": "smL/min",
                    "DryCal Avg. smL/min": "smL/min",
                    "Temp. Deg C": "degC",
                    "Pressure mBar": "mbar",
                },
            },
        ),
    ],
)
def test_basic_csv(infile, params, datadir):
    os.chdir(datadir)
    ret = extract(
        fn=infile,
        parameters=Basic_csv(filetype="basic.csv", parameters={**params}).parameters,
        encoding="utf8",
        locale="en_GB",
        timezone="Europe/Berlin",
    )
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)


@pytest.mark.parametrize(
    "infile, params, locale",
    [
        (
            "sheet.UK.tsv",
            {
                "sep": "\t",
                "timestamp": {
                    "date": {"index": 0, "format": "%d/%m/%Y"},
                    "time": {"index": 1, "format": "%H:%M:%S"},
                },
            },
            "en_GB.UTF-8",
        ),
        (
            "sheet.DE.tsv",
            {
                "sep": "\t",
                "timestamp": {
                    "date": {"index": 0, "format": "%d.%m.%Y"},
                    "time": {"index": 1, "format": "%H:%M:%S"},
                },
            },
            "de_DE.UTF-8",
        ),
        (
            "sheet.US.tsv",
            {
                "sep": "\t",
                "timestamp": {
                    "date": {"index": 0, "format": "%m/%d/%Y"},
                    "time": {"index": 1, "format": "%H:%M:%S %p"},
                },
            },
            "en_US.UTF-8",
        ),
    ],
)
def test_basic_csv_locale(infile, params, locale, datadir):
    os.chdir(datadir)
    ret = extract(
        fn=infile,
        parameters=Basic_csv(filetype="basic.csv", parameters={**params}).parameters,
        encoding="utf8",
        locale=locale,
        timezone="Europe/Berlin",
    )
    outfile = "sheet.XX.tsv.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)


@pytest.mark.parametrize(
    "infile, params, encoding",
    [
        (
            "log 2021-09-17 11-26-14.140.csv",
            {
                "timestamp": {
                    "timestamp": {"index": 0, "format": '"%Y-%m-%d %H:%M:%S.%f"'}
                },
                "units": {},
            },
            "utf-8-sig",
        ),
    ],
)
def test_basic_csv_encoding(infile, params, encoding, datadir):
    os.chdir(datadir)
    ret = extract(
        fn=infile,
        parameters=Basic_csv(filetype="basic.csv", parameters={**params}).parameters,
        encoding=encoding,
        locale="en_GB",
        timezone="Europe/Berlin",
    )
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)



import importlib
import logging
import json
from datatree import DataTree
from yadg import dgutils
from dgbowl_schemas.yadg.dataschema import ExtractorFactory, FileType


logger = logging.getLogger(__name__)


def extract(
    filetype: str,
    path: str,
    timezone: str = None,
    encoding: str = None,
    locale: str = None,
    **kwargs: dict,
) -> DataTree:
    """
    Extract data and metadata from a path using the supplied filetype.

    A wrapper around the :func:`extract_from_path` worker function, which creates a
    default extractor object.

    Parameters
    ----------

    filetype:
        Specifies the filetype. Has to be a filetype supported by the dataschema.

    path:
        A :class:`pathlib.Path` object pointing to the file to be extracted.

    timezone:
        A :class:`str` containing the TZ identifier, e.g. "Europe/Berlin".

    encoding:
        A :class:`str` containing the encoding, e.g. "utf-8".

    locale:
        A :class:`str` containing the locale name, e.g. "de_CH".

    """
    extractor = ExtractorFactory(
        extractor={
            "filetype": filetype,
            "timezone": timezone,
            "encoding": encoding,
            "locale": locale,
        }
    ).extractor

    return extract_from_path(path, extractor)


def extract_from_path(
    path: str,
    extractor: FileType,
) -> DataTree:
    """
    Extracts data and metadata from the provided path using the supplied extractor.

    The individual extractor functionality of yadg is called from here. The data is
    always returned as a :class:`DataTree`. The ``original_metadata`` entries in
    the returned objects are flattened using json serialisation. The returned objects
    have a :func:`to_netcdf` as well as a :func:`to_dict` method, which can be used to
    write the returned object into a file.
    """

    m = importlib.import_module(f"yadg.extractors.{extractor.filetype}")
    func = getattr(m, "extract")

    # Func should always return a datatree.DataTree
    ret: DataTree = func(fn=path, **vars(extractor))
    jsonize_orig_meta(ret)

    ret.attrs.update(
        {
            "yadg_provenance": "yadg extract",
            "yadg_extract_date": dgutils.now(asstr=True),
            "yadg_extract_filename": str(path),
            "yadg_extract_Extractor": extractor.model_dump_json(exclude_none=True),
        }
    )
    ret.attrs.update(dgutils.get_yadg_metadata())

    return ret


def jsonize_orig_meta(obj: DataTree):
    for k in obj:
        if isinstance(obj[k], DataTree):
            jsonize_orig_meta(obj[k])
    if "original_metadata" in obj.attrs:
        obj.attrs["original_metadata"] = json.dumps(obj.attrs["original_metadata"])
    if "fulldate" in obj.attrs:
        obj.attrs["fulldate"] = int(obj.attrs["fulldate"])


__all__ = ["extract"]



import os
import logging
import json
import yaml
import shutil
import hashlib
from pathlib import Path
from datatree import DataTree
from xarray import Dataset

from dgbowl_schemas.yadg import to_dataschema
from yadg import core, dgutils, extractors


logger = logging.getLogger(__name__)


def _load_file(infile: str) -> dict:
    with open(infile, "r") as inf:
        if infile.endswith("json"):
            obj = json.load(inf)
        elif infile.endswith("yml") or infile.endswith("yaml"):
            obj = yaml.safe_load(inf)
        else:
            logging.critical("Filename type not recognised: '%s'", infile)
            raise RuntimeError(f"Filename type not recognised: '{infile}'")
    return obj


def _zip_file(folder: str, outpath: str, method: str = "zip") -> str:
    if method in {"zip", "tar"}:
        fn = f"{outpath}.{method}"
    elif method in {"bztar", "gztar", "xztar"}:
        fn = f"{outpath}.tar.{method[:2]}"
        if method == "bztar":
            fn = f"{fn}2"
    logger.debug("Archiving.")
    shutil.make_archive(outpath, method, root_dir=folder)
    logger.debug("Hashing.")
    m = hashlib.sha1()
    with open(fn, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            m.update(chunk)
    return fn, m.hexdigest()


def _obj_to_meta_dict(dt: DataTree) -> dict:
    ret = {}
    for k, v in dt.to_dict().items():
        if isinstance(v, Dataset):
            ret[k] = v.to_dict(data=False)
        elif isinstance(v, DataTree):
            ret[k] = _obj_to_meta_dict(v)
        else:
            raise RuntimeError(f"Object {k!r} is type {type(v)!r}.")
    return ret


def process(
    *,
    infile: str,
    outfile: str,
    ignore_merge_errors: bool,
    **kwargs: dict,
) -> None:
    """
    The ``process`` subcommand of yadg.

    This function first checks that the supplied ``infile`` exists, is a valid
    dataschema, and if yes, proceeds to process the dataschema into a datatree. If
    this is successful, the datatree is written out into ``outfile`` (which is
    ``"datagram.nc"`` by default).

    """
    assert os.path.exists(infile) and os.path.isfile(infile), (
        f"Supplied dataschema filename '{infile}' does not exist "
        "or is not a valid file."
    )

    logger.info(f"Reading input file from '{infile}'.")
    schema = _load_file(infile)

    logger.info("Loading dataschema.")
    inobj = to_dataschema(**schema)
    ds = dgutils.update_schema(inobj)

    logger.debug("Processing dataschema")
    datagram = core.process_schema(ds, strict_merge=not ignore_merge_errors)

    logger.info("Saving datatree to '%s'.", outfile)
    datagram.to_netcdf(outfile, engine="h5netcdf")


def update(
    *,
    infile: str,
    outfile: str,
    **kwargs: dict,
) -> None:
    """
    The ``update`` subcommand of yadg.

    This function updates the dataschema present in the ``infile`` to comply with the
    newest version of dataschema, and saves the resulting object into ``outfile`` (which
    is the ``infile`` with  a `".new.json"` suffix by default).

    """
    assert os.path.exists(infile) and os.path.isfile(infile), (
        f"Supplied dataschema filename '{infile}' does not exist "
        "or is not a valid file."
    )

    logger.info("Reading input file from '%s'.", infile)
    inobj = _load_file(infile)

    if outfile is None:
        name, ext = os.path.splitext(infile)
        outfile = f"{name}.new.json"

    outobj = dgutils.update_schema(inobj)

    logger.info("Writing new dataschema into '%s'.", outfile)
    with open(outfile, "w") as out:
        json.dump(outobj.model_dump(), out, indent=1)


def preset(
    *,
    preset: str,
    folder: str,
    outfile: str,
    process: bool,
    archive: bool,
    packwith: str,
    ignore_merge_errors: bool,
    **kwargs: dict,
) -> None:
    """
    The ``preset`` subcommand of yadg.

    If ``folder`` is an existing folder and ``preset`` a valid dataschema template file,
    this function will prepend the specified ``folder`` to all relative paths in the
    ``preset``, converting them to absolute paths.

    The resulting dataschema will be saved in the supplied ``outfile``.

    Alternatively, if ``process`` is specified, the created dataschema will be
    directly processed into a datatree, which is then saved in ``outfile``.

    Additionally, the contents of the ``folder`` can be archived (if ``archive`` is
    set), using a compression algorithm of your choice.

    """
    assert os.path.exists(folder) and os.path.isdir(folder), (
        f"Supplied folder path '{folder}' does not exist " "or is not a valid folder."
    )

    if not os.path.isabs(folder) and not process:
        logger.warning(
            f"The provided path '{folder}' is a relative path. The generated schema "
            "likely will not work outside current working directory."
        )

    assert os.path.exists(preset) and os.path.isfile(preset), (
        f"Supplied preset path '{preset}' does not exist " "or is not a valid file."
    )

    logger.info("Reading input file from '%s'.", preset)
    preset = _load_file(preset)

    logger.info("Processing preset as schema.")
    schema = to_dataschema(**preset)

    logger.info("Creating a schema from preset for '%s'.", folder)
    ds = dgutils.update_schema(dgutils.schema_from_preset(schema, folder))

    logger.info("Loaded dataschema version '%s'", ds.version)
    if process:
        logger.info("Processing created schema.")
        datagram = core.process_schema(ds, strict_merge=not ignore_merge_errors)
        outfile = "datagram.nc" if outfile is None else outfile
        if archive:
            zipfile = outfile.replace(".nc", "")
            logger.info("Zipping input folder into '%s'", zipfile)
            fn, hash = _zip_file(folder, zipfile, method=packwith)
            datagram.attrs["data_archive_sha-1"] = hash
            datagram.attrs["data_archive_path"] = fn
        logger.info("Saving datagram to '%s'.", outfile)
        datagram.to_netcdf(outfile, engine="h5netcdf")
    else:
        if archive:
            logger.warning(
                "The --archive option must be supplied along with --process. "
                "Continuing without archiving."
            )
        outfile = "schema.json" if outfile is None else outfile
        logger.info("Saving schema to '%s'.", outfile)
        with open(outfile, "w") as ofile:
            json.dump(ds.model_dump(), ofile, indent=1)


def extract(
    *,
    filetype: str,
    infile: str,
    outfile: str,
    meta_only: bool,
    **kwargs: dict,
) -> None:
    """
    The ``extract`` subcommand of yadg.

    If ``filetype`` is known to yadg, it will attempt to extract all data from the
    provided ``infile``.

    The data is returned as a :class:`xarray.Dataset` or a datatree, and is stored in
    a NetCDF file. The output location can be configured using the ``outfile``
    argument, by default this is set to the stem of ``infile`` with a ``.nc`` suffix.

    Optionally, an export of just the metadata can be requested by setting the
    ``meta_only`` argument, in this case the output is a json file.

    """
    path = Path(infile)

    assert path.is_file(), (
        f"Supplied object filename '{infile}' does not exist " "or is not a valid file."
    )

    if outfile is None:
        outpath = path.with_suffix(".json" if meta_only else ".nc")
    else:
        outpath = Path(outfile)

    ret = extractors.extract(filetype, path, **kwargs)
    if meta_only:
        meta = _obj_to_meta_dict(ret)
        with outpath.open("w", encoding="UTF-8") as target:
            json.dump(meta, target)
    else:
        ret.to_netcdf(outpath, engine="h5netcdf")



import logging
from datatree import DataTree

from dgbowl_schemas.yadg.dataschema import DataSchema
from yadg import dgutils
from yadg.extractors import extract_from_path

logger = logging.getLogger(__name__)


def process_schema(dataschema: DataSchema, strict_merge: bool = False) -> DataTree:
    """
    The main processing function of yadg.

    Takes in a :class:`DataSchema` object and returns a single :class:`DataTree` created
    from the :class:`DataSchema`.

    """
    if strict_merge:
        concatmode = "identical"
    else:
        concatmode = "drop_conflicts"

    while hasattr(dataschema, "update"):
        dataschema = dataschema.update()

    root = DataTree()
    root.attrs = {
        "yadg_provenance": "yadg process",
        "yadg_process_date": dgutils.now(asstr=True),
        "yadg_process_DataSchema": dataschema.model_dump_json(),
    }
    root.attrs.update(dgutils.get_yadg_metadata())

    for si, step in enumerate(dataschema.steps):
        logger.info(f"Processing step {si}.")

        # Backfill default timezone, locale, encoding.
        if step.extractor.timezone is None:
            step.extractor.timezone = dataschema.step_defaults.timezone

        if step.extractor.locale is None:
            step.extractor.locale = dataschema.step_defaults.locale
        if step.extractor.encoding is None:
            step.extractor.encoding = dataschema.step_defaults.encoding

        if step.tag is None:
            step.tag = f"{si}"

        todofiles = step.input.paths()
        vals = None
        if len(todofiles) == 0:
            logger.warning(f"No files processed by step '{step.tag}'.")
            vals = {}
        for tf in todofiles:
            logger.info(f"Processing file '{tf}'.")
            tasks = extract_from_path(path=tf, extractor=step.extractor).to_dict()
            fvals = {}
            for name, dset in tasks.items():
                # The root datatree node may sometimes carry metadata, even if
                # there are no variables - we don't add 'uts' to those.
                if name == "/" and len(dset.variables) == 0:
                    fvals[name] = dset
                # Otherwise, we want to process any 'step.externaldate' commands
                else:
                    fvals[name] = dgutils.complete_uts(
                        dset, tf, step.externaldate, step.extractor.timezone
                    )
                # Remove metadata entries we know will differ between different files.
                for k in {"yadg_extract_date", "yadg_extract_filename"}:
                    if k in fvals[name].attrs:
                        del fvals[name].attrs[k]

            vals = dgutils.merge_dicttrees(vals, fvals, concatmode)

        stepdt = DataTree.from_dict({} if vals is None else vals)
        stepdt.name = step.tag
        stepdt.parent = root
    return root



"""
Module for parsing PicoTech PicoLog files, such as those generated by the TC-08
thermocouple reader.

Usage
`````
Available since ``yadg-5.1``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.PicoLog_tc08

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float           # Unix timestamp, optional
      data_vars:
        {{ name }}      (uts)             # Temperature of the named thermocouple

Metadata
````````
Metadata about the TC-08 device is read from the overall metadata file. The following
parameters are currently parsed:

- ``Model``: model number of the network analyseer
- ``Serial``: serial number of the device
- ``Software Version``: software version with which the file was created


Uncertainties
`````````````
Uncertainties in temperatures are set to 2.2°C, which is the usual value for a Type-K
thermocouple.

Notes on file structure
```````````````````````
The ``.picolog`` files are actually ``.tar`` archives, which include the following
files:

- a ``metadata.json`` file, containing information about the device and the
  thermocouples plugged in,
- a ``data-map.json.gz`` file, containing information about chunks of data corresponding
  to each thermocouple,
- a set of zipped binary files with data averaged at different time resolutions

Currently, the data is extracted from the ``1.0.gz`` file only.

.. codeauthor::
    Peter Kraus

"""

import os
import json
import numpy as np
import tarfile
import gzip
import tempfile
import xarray as xr
from datatree import DataTree

from yadg import dgutils


def extract(
    *,
    fn: str,
    **kwargs: dict,
) -> DataTree:
    with tarfile.open(fn, mode="r") as tf:
        with tempfile.TemporaryDirectory() as tempdir:
            tf.extractall(tempdir)

            # Get file metadata
            with open(os.path.join(tempdir, "metadata.json")) as inf:
                metadata = json.load(inf)

            # Get data from 1.0.gz
            with gzip.open(os.path.join(tempdir, "data-map.json.gz"), "rb") as inp:
                dmap = json.loads(inp.read())
            with gzip.open(os.path.join(tempdir, "1.0.gz"), "rb") as inp:
                raw = inp.read()

    uts = dgutils.str_to_uts(timestamp=metadata["startDate"], timezone=None)

    # Here we assume that "items" only has one element
    devdata = metadata["devices"]["items"][0]
    attrs = {
        "Software Version": metadata["appVersion"],
        "Model": devdata["__type"],
        "Serial": devdata["serialInternal"],
    }

    # Get necessary device metadata from the file metadata
    devices = {}
    for k, v in devdata.items():
        if isinstance(v, dict) and v["enabled"]:
            devices[k] = {
                "id": v["channelIdentifier"],
                "type": v["channelType"],
                "name": v["description"],
                "xmul": v["captureConfig"]["interval"] // 1000,
                "npts": v["capturedSamples"],
            }

    # Convert bytes to floats
    points = {}
    for tag, params in dmap:
        _, __, namestr = tag.split(".")
        id, res, time = namestr.split("/")
        archive, start, length = params
        if id not in points:
            points[id] = np.empty(0, dtype=">f4")
        if res == "1" and archive == "1.0.gz":
            new = np.frombuffer(raw, offset=start, dtype=">f4", count=length // 4)
            points[id] = np.concatenate((points[id], new))
        elif res == "1":
            raise RuntimeError(f"Resolution of 1 but archive is {archive!r}.")

    # Push the data into the Dataset
    ds = xr.Dataset(coords={"uts": (["uts"], [])})
    for id, data in points.items():
        for k, meta in devices.items():
            if meta["id"] == id:
                #  The type of the device should be thermocouple
                if meta["type"] == "thermocouple":
                    unit = {"units": "degC"}
                else:
                    raise RuntimeError("Unknown type {meta['type']!r}.")
                yvals = data[~np.isnan(data)]
                ydevs = np.ones(len(yvals)) * 2.2
                xvals = np.arange(len(yvals)) * meta["xmul"] + uts
                newds = xr.Dataset(
                    data_vars={
                        meta["name"]: (["uts"], yvals, unit),
                        f"{meta['name']}_std_err": (["uts"], ydevs, unit),
                    },
                    coords={"uts": (["uts"], xvals)},
                )
                ds = xr.merge((ds, newds))
    for var in ds.variables:
        if f"{var}_std_err" in ds.variables:
            ds[var].attrs["ancillary_variables"] = f"{var}_std_err"
        elif var.endswith("_std_err"):
            end = var.index("_std_err")
            if var[:end] in ds.variables:
                ds[var].attrs["standard_name"] = f"{var[:end]} standard_error"

    ds.attrs = dict(original_metadata=attrs)
    return DataTree(ds)



import pytest
import os
import json
import yadg.dgutils
import yadg.core


def standard_datagram_test(datagram, testspec):
    assert len(datagram.children) == testspec["nsteps"], "wrong number of steps"
    if isinstance(testspec["step"], str):
        step = datagram[f"{testspec['step']}"]
    else:
        name = list(datagram.children.keys())[testspec["step"]]
        step = datagram[name]
    if len(step.children) > 0:
        for k, v in step.items():
            assert len(v["uts"]) == testspec["nrows"], (
                f"wrong number of timesteps in a child Dataset {k}: "
                f"ret: {len(v['uts'])}, ref: {testspec['nrows']}"
            )
    if "uts" in step:
        assert len(step["uts"]) == testspec["nrows"], (
            "wrong number of timesteps in a step: "
            f"ret: {len(step['uts'])}, ref: {testspec['nrows']}"
        )
    datagram.to_netcdf(".test.nc", engine="h5netcdf")


@pytest.mark.parametrize(
    "input, ts",
    [
        ("data_1", {"nsteps": 2, "step": 0, "nrows": 1662}),
        ("data_2", {"nsteps": 2, "step": 0, "nrows": 2}),
    ],
)
def test_preset(input, ts, datadir):
    os.chdir(datadir)

    with open(f"{input}.preset.json") as infile:
        preset = json.load(infile)

    schema = yadg.dgutils.update_schema(preset)
    ds = yadg.dgutils.schema_from_preset(schema, input)
    ret = yadg.core.process_schema(ds)
    standard_datagram_test(ret, ts)
